unit COMObjContextMenu;
{..$DEFINE Debug}
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   2 February 1999
   Updated 17 February 1999
    1999 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use the 
   component or any part of its code, you must include an 
   acknowledgement in the finished code or executable (in the About 
   box and/or the Help file).

   TCOMObjContextMenu:
   Generic TCOMObjShellExtInit descendant that implements the
   IContextMenu, IContextMenu2 and IContextMenu3 interfaces.
   It is designed so the user need only override the
   DoQuery, DoMenuChar, DoInitMenuPopup, DoMeasureItem and DoDrawItem methods.

   Context menu handlers should be registered under the target file type.

   For example,
   HKEY_CLASSES_ROOT\.del = DelShellFile
   HKEY_CLASSES_ROOT\DelShellFile\shellex\ContextMenuHandlers\DUTShellContextMenu = CLSID_DelContextMenu

   ContextMenu Handlers can also be used as Find Extensions, which appear under the Start Menu's
Find entry and Windows Explorer's Tools:Find menu.
   Static Find Extensions are registered under:
   HKEY_LOCAL_MACHINE\Software\Microsoft\Windows\CurrentVersion\Explorer\FindExtensions\Static
   Each static find extension creates a sub-key with a unique (non-significant) name, whose
value is the extension's CLSID, like so:
   HKEY_LOCAL_MACHINE\...\Static\MyFind = "<clsid>"
   A surefire way of creating a unique sub-key name is to use the CLSID.  It is not very
human-readable, however.
   Each command the extension adds to the Find menu is registered sequentially as sub-keys "0",
"1", "2" and so on, the value of each sub-key being the menu string.  Further, each menu
entry sub-key may have a DefaultIcon subkey specifying the small icon to display in the menu
beside the menu string.  Like so:
   HKEY_LOCAL_MACHINE\...\Static\MyFind\0 = "&My Find Utility"
   HKEY_LOCAL_MACHINE\...\Static\MyFind\0\DefaultIcon = "c:\MyFind.dll,0"
   HKEY_LOCAL_MACHINE\...\Static\MyFind\1 = "&My Other Find Utility"
   HKEY_LOCAL_MACHINE\...\Static\MyFind\1\DefaultIcon = "c:\MyFind.dll,1"
   If the user selects one of your Static Find Extension menu entries, the browser loads the
context menu handler specified by the CLSID and calls IContextMenu::InvokeCommand.  The
low-order word of lpici.lpVerb contains the numeric representation of the command key name
(i.e. zero or one in our example).  In response to the command, the context menu handler
must launch the find utility.
   (It is unclear whether Initialize is skipped; seems unlikely)

   A dynamic Find Extension is loaded when the shell loads at startup, and is kept loaded
until the shell exits.  You add a dynamic Find extension by creating a registry entry under:
   HKEY_LOCAL_MACHINE\Software\Microsoft\Windows\CurrentVersion\Explorer\FindExtensions
   You create a unique sub-key under this key.  The sub-key name is not significant as long
as it is unique among the other key names at the same level.  The default value for this key
is the CLSID string of the context menu handler.
   HKEY_LOCAL_MACHINE\...\FindExtensions\MyFind = "<clsid>"
   Dynamic Find Extensions are loaded normally except that the pidlFolder and lpdobj arguments
in the extension's IShellExtInit::Initialize will be NULL.  The rest of the behavior is as per
a normal context menu extension.  Menu icons are not supported for dynamic Find extensions.
*)
interface

uses
   Windows,
   Classes, { TPersistent }
   Graphics, { Graphics.TBitmap }
// ShlObj, { IContextMenu }
   COMObjShellExtInit, { TCOMObjShellExtInit }
   DUTBits; { TDUTBits }

const
   { IContextMenu3 isn't declared at all in ShlObj.pas }
   IID_IContextMenu3 : TGUID = (
   D1:$BCFCE0A0; D2:$EC17; D3:$11D0; D4:($8D,$10,$00,$A0,$C9,$0F,$27,$19));
   SID_IContextMenu3 = '{BCFCE0A0-EC17-11d0-8D10-00A0C90F2719}';
//
   IID_IContextMenu: TGUID = (
   D1:$000214E4; D2:$0000; D3:$0000; D4:($C0,$00,$00,$00,$00,$00,$00,$46));
   SID_IContextMenu       = '{000214E4-0000-0000-C000-000000000046}';
   IID_IContextMenu2: TGUID = (
   D1:$000214F4; D2:$0000; D3:$0000; D4:($C0,$00,$00,$00,$00,$00,$00,$46));
   SID_IContextMenu2      = '{000214F4-0000-0000-C000-000000000046}';
//
type
//
   PCMInvokeCommandInfo = ^TCMInvokeCommandInfo;
   TCMInvokeCommandInfo = packed record
      cbSize       : DWORD;   { must be sizeof(CMINVOKECOMMANDINFO) }
      fMask        : DWORD;   { any combination of CMIC_MASK_* }
      hwnd         : HWND;    { might be NULL (indicating no owner window) }
      lpVerb       : LPCSTR;  { either a string of MAKEINTRESOURCE(idOffset) }
      lpParameters : LPCSTR;  { might be NULL (indicating no parameter) }
      lpDirectory  : LPCSTR;  { might be NULL (indicating no specific directory) }
      nShow        : Integer; { one of SW_ values for ShowWindow() API }
      dwHotKey     : DWORD;
      hIcon        : THandle;
   end; { TCMInvokeCommandInfo }
   IContextMenu = interface(IUnknown)
      [SID_IContextMenu]
      function QueryContextMenu(Menu : HMENU; indexMenu, idCmdFirst, idCmdLast, uFlags : UINT) : HResult; stdcall;
      function InvokeCommand(var lpici : TCMInvokeCommandInfo) : HResult; stdcall;
      function GetCommandString(idCmd, uType : UINT; pwReserved : PUINT; pszName : LPSTR; cchMax : UINT) : HResult; stdcall;
   end; { IContextMenu }
//
   { IContextMenu2 is incorrectly declared in ShlObj.pas }
   IContextMenu2 = interface(IContextMenu)
      [SID_IContextMenu2]
      function HandleMenuMsg(uMsg : UINT; wParam : WPARAM; lParam : LPARAM) : HResult; stdcall;
   end; { IContextMenu2 }

   { IContextMenu3 isn't declared at all in ShlObj.pas }
   IContextMenu3 = interface(IContextMenu2)
      [SID_IContextMenu3]
      function HandleMenuMsg2(uMsg : UINT; wParam : WPARAM; lParam : LPARAM; var MsgResult : LongInt) : HResult; stdcall;
   end; { IContextMenu3 }

   TIntList = class(TList)
   {
   A simple TList descendant that handles the contained "pointers" as LongInts
   }
   protected
   { Available to descendants }
      function Get(Index : Integer) : Integer;
      procedure Put(Index, Item : Integer);
   public
   { Available at run-time  }
      function Add(Item : Integer) : Integer;
      function IndexOf(Item : Integer): Integer;
      procedure Insert(Index, Item : Integer);
      function Remove(Item : Integer) : Integer;
      property Items[Index : Integer] : Integer read Get write Put; default;
   end; { TIntList }

   TContextMenuStringType = (cmstCaption, cmstHint, cmstVerb);

   TContextMenu = class(TPersistent)
   {
   TContextMenu is a TMenuItem-compatible TPersistent descendant that
   holds those properties required by TCOMObjContextMenu.
   Captions[] gives the menu item strings;
   Checked[] states whether the menu entry is checked;
   CheckedBitmaps[] gives the handles of the (monochrome) bitmaps to use as check marks;
      (You can use polychrome bitmaps but the results may not be what you expect)
      The property owns the handle once it is assigned to it, and will free it for you.
      You can use GetSystemMetrics with CXMENUCHECK and CYMENUCHECK to obtain the
      checkmark dimensions beforehand.
      Note that this property expects a Windows HBITMAP, so you can use the Windows API
      without needing to go through Delphi's Graphics.TBitmap; you can also use a
      Graphics.TBitmap's Handle property but you'll then need to Create and Free the
      object in the context menu object's unit's initialization and finalization blocks.
      If you Free the object after passing its Handle to CheckedBitmaps, the context menu
      will NOT be able to fetch the icon (the default checkmark will then be used instead).
      The utility function DUTUtilsGraphics.DuplicateBitmap is supplied for this purpose; use it
      to create a duplicate hBitmap that the ContextMenu can safely use and free on its own.
   Grayed[] states whether the menu entry is grayed out;
   HasCheckedBitmap[] states whether the menu entry has a checked bitmap set;
   HasMenuBitmap[] states whether the menu entry uses a bitmap instead of a caption;
   HasSeparator[] states whether the menu entry is followed by a separator;
   HasSubmenu[] states whether the menu entry has a submenu (read-only);
   HasUncheckedBitmap[] states whether the menu entry has an unchecked bitmap set;
   Hints[] gives the menu item Help strings;
   ID[] gives the menu item's relative command ID (read-only);
      (Invisible entries won't have an ID and return -1)
   IDOffset is the offset to apply to the above IDs to convert them to absolute IDs.
   IsOwnerDrawn[] states whether the menu entry is owner-drawn;
   MenuBitmaps[] gives the handles of the bitmaps to use instead of Captions;
      Here too the utility function DUTUtilsGraphics.DuplicateBitmap may come in handy.
   OnClick[] gives the event handler for each command;
   OwnerDrawnData[] is the user-specified 32-bit "data" associated with the menu entry;
      Note that if MenuBitmaps[] is also set for that entry, it takes precedence
      A zero value in this array means the entry is NOT owner-drawn.
   Submenu[] gives the sub-TContextMenu, if any;
      (To create a submenu, create a new TContextMenu, fill it in,
      then set it as its parent's submenu (at the proper parent index)
      You need only Free the parent to Free all the submenus it owns)
   UncheckedBitmaps[] gives the handles of the (monochrome) bitmaps to use as "uncheck marks";
      (You can use polychrome bitmaps but the results may not be what you expect)
      The property owns the handle once it is assigned to it, and will free it for you.
   Visible[] states whether the menu entry is visible;
   Verbs[] gives the menu item language-independent Command Strings.

   All these arrays (except Visible) also have a "byID" version that
   looks up the value based on its assigned relative ID.
   }
   private
   { Unavailable }
      FIDOffset,                //The system-supplied offset to apply to IDs.  Used by InsertInHMenu only.
      FRootID    : Integer;     //The internal ID offset to apply to IDs.
      {
      Notes:
      Internally, the RootID is the offset of the TContextMenu's IDs with respect
      to its owner TContextMenu.  This keeps the entire FContextMenu IDs within
      the span registered by QueryContextMenu.  It is a RELATIVE offset.
      The IDOffset property is used by InsertInHMenu to express an ABSOLUTE offset,
      as specified by the QueryContextMenu's idCmdFirst.  It needs to be set only
      for FContextMenu; the InsertInHMenu call will set the submenus' IDOffsets.
      Remember that the GetCommandString and InvokeCommand calls pass in a relative
      ID offset and that the HandleMenuMsg calls pass in an absolute ID.
      }
      FCaptions,
      FHints,
      FVerbs     : TStringList;
      //ID will be stored in FHints.Objects as a TIntegerObject
      //OnClick will be stored in FVerbs.Objects as a TNotifyObject
      //Submenus will be stored in FCaptions.Objects.
      FMenuBitmaps,
      FCheckedBitmaps,
      FUncheckedBitmaps,
      FOwnerDrawnData : TIntList;
      FChecked,
      FGrayed,
      FHasSeparator,
      FVisible          : TDUTBits;
   protected
   { Available to descendants  }
      function GetCaption(Index : Integer) : String; virtual;
      function GetCaptionByID(CommandID : Integer) : String; virtual;
      function GetChecked(Index : Integer) : Boolean; virtual;
      function GetCheckedByID(CommandID : Integer) : Boolean; virtual;
      function GetCheckedBitmap(Index : Integer) : HBITMAP; virtual;
      function GetCheckedBitmapByID(CommandID : Integer) : HBITMAP; virtual;
      function GetCount : Integer; virtual;
      function GetGrayed(Index : Integer) : Boolean; virtual;
      function GetGrayedByID(CommandID : Integer) : Boolean; virtual;
      function GetHasCheckedBitmap(Index : Integer) : Boolean; virtual;
      function GetHasCheckedBitmapByID(CommandID : Integer) : Boolean; virtual;
      function GetHasMenuBitmap(Index : Integer) : Boolean; virtual;
      function GetHasMenuBitmapByID(CommandID : Integer) : Boolean; virtual;
      function GetHasSeparator(Index : Integer) : Boolean; virtual;
      function GetHasSeparatorByID(CommandID : Integer) : Boolean; virtual;
      function GetHasSubmenu(Index : Integer) : Boolean; virtual;
      function GetHasSubmenuByID(CommandID : Integer) : Boolean; virtual;
      function GetHasUncheckedBitmap(Index : Integer) : Boolean; virtual;
      function GetHasUncheckedBitmapByID(CommandID : Integer) : Boolean; virtual;
      function GetHint(Index : Integer) : String; virtual;
      function GetHintByID(CommandID : Integer) : String; virtual;
      function GetID(Index : Integer) : Integer; virtual;
      function GetIDCount : Integer; virtual;
      function GetIsOwnerDrawn(Index : Integer) : Boolean; virtual;
      function GetIsOwnerDrawnByID(CommandID : Integer) : Boolean; virtual;
      function GetMenuBitmap(Index : Integer) : HBITMAP; virtual;
      function GetMenuBitmapByID(CommandID : Integer) : HBITMAP; virtual;
      function GetOnClick(Index : Integer) : TNotifyEvent; virtual;
      function GetOnClickByID(CommandID : Integer) : TNotifyEvent; virtual;
      function GetOwnerDrawnData(Index : Integer) : Integer; virtual;
      function GetOwnerDrawnDataByID(CommandID : Integer) : Integer; virtual;
      function GetSubmenu(Index : Integer) : TContextMenu; virtual;
      function GetSubmenuByID(CommandID : Integer) : TContextMenu; virtual;
      function GetTotalCount : Integer; virtual;
      function GetUncheckedBitmap(Index : Integer) : HBITMAP; virtual;
      function GetUncheckedBitmapByID(CommandID : Integer) : HBITMAP; virtual;
      function GetVerb(Index : Integer) : String; virtual;
      function GetVerbByID(CommandID : Integer) : String; virtual;
      function GetVisible(Index : Integer) : Boolean; virtual;
      function GetVisibleByID(CommandID : Integer) : Boolean; virtual;
      procedure SetCaption(Index : Integer; Value : String); virtual;
      procedure SetCaptionByID(CommandID : Integer; Value : String); virtual;
      procedure SetChecked(Index : Integer; Value : Boolean); virtual;
      procedure SetCheckedByID(CommandID : Integer; Value : Boolean); virtual;
      procedure SetCheckedBitmap(Index : Integer; Value : HBITMAP); virtual;
      procedure SetCheckedBitmapByID(CommandID : Integer; Value : HBITMAP); virtual;
      procedure SetGrayed(Index : Integer; Value : Boolean); virtual;
      procedure SetGrayedByID(CommandID : Integer; Value : Boolean); virtual;
      procedure SetHasSeparator(Index : Integer; Value : Boolean); virtual;
      procedure SetHasSeparatorByID(CommandID : Integer; Value : Boolean); virtual;
      procedure SetHint(Index : Integer; Value : String); virtual;
      procedure SetHintByID(CommandID : Integer; Value : String); virtual;
      procedure SetID(Index : Integer; Value : Integer); virtual;
      procedure SetMenuBitmap(Index : Integer; Value : HBITMAP); virtual;
      procedure SetMenuBitmapByID(CommandID : Integer; Value : HBITMAP); virtual;
      procedure SetOnClick(Index : Integer; Value : TNotifyEvent); virtual;
      procedure SetOnClickByID(CommandID : Integer; Value : TNotifyEvent); virtual;
      procedure SetOwnerDrawnData(Index, Value : Integer); virtual;
      procedure SetOwnerDrawnDataByID(CommandID, Value : Integer); virtual;
      procedure SetSubmenu(Index : Integer; Value : TContextMenu); virtual;
      procedure SetSubmenuByID(CommandID : Integer; Value : TContextMenu); virtual;
      procedure SetUncheckedBitmap(Index : Integer; Value : HBITMAP); virtual;
      procedure SetUncheckedBitmapByID(CommandID : Integer; Value : HBITMAP); virtual;
      procedure SetVerb(Index : Integer; Value : String); virtual;
      procedure SetVerbByID(CommandID : Integer; Value : String); virtual;
      procedure SetVisible(Index : Integer; Value : Boolean); virtual;
      procedure SetVisibleByID(CommandID : Integer; Value : Boolean); virtual;
      procedure ResetIDs; virtual;
      procedure ZapIDs; virtual;

      property RootID : Integer
         read FRootID
         write FRootID;
   public
   { Available at run-time }
      constructor Create; virtual;
      destructor Destroy; override;
      function Add(const Verb, Caption, Hint : string; const OnClick : TNotifyEvent; const Checked, Grayed, Separator, Visible : Boolean) : Integer; virtual;
      procedure Assign(Source : TPersistent); override;
      procedure AssignTo(Dest : TPersistent); override;
      procedure Clear; virtual;
      function ContextMenuOfID(const CommandID : Integer) : TContextMenu; virtual;
      procedure Delete(Index : Integer); virtual;
      procedure Exchange(Index1, Index2 : Integer); virtual;
      function IDOfString(const AString : String; const StringType : TContextMenuStringType) : Integer; virtual;
      function IndexOfCaption(const Caption : string) : Integer; virtual;
      function IndexOfID(const CommandID : Integer) : Integer; virtual;
      function IndexOfVerb(const Verb : string) : Integer; virtual;
      function IndexOfHint(const Hint : string) : Integer; virtual;
      procedure Insert(Index : Integer; const Verb, Caption, Hint : string; const OnClick : TNotifyEvent; const Checked, Grayed, Separator, Visible : Boolean); virtual;
      function InsertInHMenu(var Menu : HMENU; var IndexOffset : Integer) : Boolean; virtual;

      property Captions[Index : Integer] : String
         read GetCaption
         write SetCaption;
      property CaptionsByID[CommandID : Integer] : String
         read GetCaptionByID
         write SetCaptionByID;
      property Checked[Index : Integer] : Boolean
         read GetChecked
         write SetChecked;
      property CheckedByID[CommandID : Integer] : Boolean
         read GetCheckedByID
         write SetCheckedByID;
      property CheckedBitmaps[Index : Integer] : HBITMAP
         read GetCheckedBitmap
         write SetCheckedBitmap;
      property CheckedBitmapsByID[CommandID : Integer] : HBITMAP
         read GetCheckedBitmapByID
         write SetCheckedBitmapByID;
      property Count : Integer
         read GetCount;
      property Grayed[Index : Integer] : Boolean
         read GetGrayed
         write SetGrayed;
      property GrayedByID[CommandID : Integer] : Boolean
         read GetGrayedByID
         write SetGrayedByID;
      property HasCheckedBitmap[Index : Integer] : Boolean
         read GetHasCheckedBitmap;
      property HasCheckedBitmapByID[CommandID : Integer] : Boolean
         read GetHasCheckedBitmapByID;
      property HasMenuBitmap[Index : Integer] : Boolean
         read GetHasMenuBitmap;
      property HasMenuBitmapByID[CommandID : Integer] : Boolean
         read GetHasMenuBitmapByID;
      property HasSeparator[Index : Integer] : Boolean
         read GetHasSeparator
         write SetHasSeparator;
      property HasSeparatorByID[CommandID : Integer] : Boolean
         read GetHasSeparatorByID
         write SetHasSeparatorByID;
      property HasSubmenu[Index : Integer] : Boolean
         read GetHasSubmenu;
      property HasSubmenuByID[CommandID : Integer] : Boolean
         read GetHasSubmenuByID;
      property HasUncheckedBitmap[Index : Integer] : Boolean
         read GetHasUncheckedBitmap;
      property HasUncheckedBitmapByID[CommandID : Integer] : Boolean
         read GetHasUncheckedBitmapByID;
      property Hints[Index : Integer] : String
         read GetHint
         write SetHint;
      property HintsByID[CommandID : Integer] : String
         read GetHintByID
         write SetHintByID;
      property ID[Index : Integer] : Integer
         read GetID;
      property IDCount : Integer
         read GetIDCount;
      property IDOffset : Integer
         read FIDOffset
         write FIDOffset;
      property IsOwnerDrawn[Index : Integer] : Boolean
         read GetIsOwnerDrawn;
      property IsOwnerDrawnByID[CommandID : Integer] : Boolean
         read GetIsOwnerDrawn;
      property MenuBitmaps[Index : Integer] : HBITMAP
         read GetMenuBitmap
         write SetMenuBitmap;
      property MenuBitmapsByID[CommandID : Integer] : HBITMAP
         read GetMenuBitmapByID
         write SetMenuBitmapByID;
      property OnClick[Index : Integer] : TNotifyEvent
         read GetOnClick
         write SetOnClick;
      property OnClickByID[CommandID : Integer] : TNotifyEvent
         read GetOnClickByID
         write SetOnClickByID;
      property OwnerDrawnData[Index : Integer] : Integer
         read GetOwnerDrawnData
         write SetOwnerDrawnData;
      property OwnerDrawnDataByID[CommandID : Integer] : Integer
         read GetOwnerDrawnDataByID
         write SetOwnerDrawnDataByID;
      property Submenu[Index : Integer] : TContextMenu
         read GetSubmenu
         write SetSubmenu;
      property SubmenuByID[CommandID : Integer] : TContextMenu
         read GetSubmenuByID
         write SetSubmenuByID;
      property TotalCount : Integer
         read GetTotalCount;
      property UncheckedBitmaps[Index : Integer] : HBITMAP
         read GetUncheckedBitmap
         write SetUncheckedBitmap;
      property UncheckedBitmapsByID[CommandID : Integer] : HBITMAP
         read GetUncheckedBitmapByID
         write SetUncheckedBitmapByID;
      property Verbs[Index : Integer] : String
         read GetVerb
         write SetVerb; default;
      property VerbsByID[CommandID : Integer] : String
         read GetVerbByID
         write SetVerbByID;
      property Visible[Index : Integer] : Boolean
         read GetVisible
         write SetVisible;
   end; { TContextMenu }

   TCOMObjContextMenu = class(TCOMObjShellExtInit, IContextMenu, IContextMenu2, IContextMenu3)
   protected
      FContextMenu : TContextMenu;
      {
      DoQuery is responsible for populating this object (q.v.).
      By default it is empty.
      }
      FhWnd : HWND;
      {
      Valid when the menu item event handlers are called, but not before.
      This is the handle to pass as the owner for any UI elements.
      }
      function DoQuery : HResult; virtual;
      {
      This function is called when the QueryContextMenu occurs;
      it is expected the FContextMenu will be populated by it.
      Should return S_OK for success, any COM error code for failure.
      The default implementation simply returns S_OK, leaving FContextMenu empty.
      }
      function DoMenuMsg(uMsg : UINT; wParam : WPARAM; lParam : LPARAM) : HResult; virtual;
      {
      This function is called when the HandleMenuMsg occurs;
      uMsg is one of WM_INITMENUPOPUP, WM_MEASUREITEM or WM_DRAWITEM messages
      (this allows the menu to handle owner-drawn items).
      It is this method's responsibility to update the FContextMenu so it remains
      in synch with the menu pointed to by the menu handle.
      Should return S_OK for success, any COM error code for failure.
      The default implementation simply returns S_OK.
      }
      procedure DoInitMenuPopup(Menu : HMENU; Index : Integer; SystemMenu : Boolean); virtual;
      {
      This method is called by HandleMenuMsg when processing WM_INITMENUPOPUP;
      Menu is the handle of the menu that is initializing;
      Index is the position index of the item in the owning menu that has the Menu as its popup;
      SystemMenu is True if the Menu is a System menu, False otherwise.
      The default implementation does nothing.
      }
      procedure DoMeasureItem(var MeasureItem : TMeasureItemStruct); virtual;
      {
      This method is called by HandleMenuMsg when processing WM_MEASUREITEM;
      the MeasureItem record is to be filled in.
      The default implementation does nothing.
      }
      procedure DoDrawItem(var DrawItem : TDrawItemStruct); virtual;
      {
      This method is called by HandleMenuMsg when processing WM_DRAWITEM.
      The default implementation does nothing.
      }
      procedure DoMenuChar(Key : Char; Menu : HMENU; var Index, Action : Integer); virtual;
      {
      This method is called by HandleMenuMsg in the WM_MENUCHAR case;
      the Index and Action values passed in are suggestions, based on matching
      the accelerator keys in the Captions.
      Key is the ASCII character that was pressed;
      Menu is the handle of the context menu;
      Index is the index of the menu item to apply the Action to;
      Action is one of:
      MNC_IGNORE     (Default) The key press is unrecognized
      MNC_CLOSE      The menu will be closed
      MNC_EXECUTE    The menu item returned will be selected and executed
      MNC_SELECT     The menu item returned will be highlighted (receive the focus)
      In your method, you may change the Index and Action.
      The default implementation does nothing.
      You should rarely need to override this.
      }
   public
   { Available at run-time }
      //IContextMenu
      function QueryContextMenu(Menu : HMENU; indexMenu, idCmdFirst, idCmdLast, uFlags : UINT) : HResult; stdcall;
      function InvokeCommand(var lpici : TCMInvokeCommandInfo) : HResult; stdcall;
      function GetCommandString(idCmd, uType : UINT; pwReserved : PUINT; pszName : LPSTR; cchMax : UINT) : HResult; stdcall;
      //IContextMenu2
      function HandleMenuMsg(uMsg : UINT; wParam : WPARAM; lParam : LPARAM) : HResult; stdcall;
      //IContextMenu3
      function HandleMenuMsg2(uMsg : UINT; wParam : WPARAM; lParam : LPARAM; var MsgResult : LongInt) : HResult; stdcall;
    end;

implementation

uses
{$IFDEF Debug}
DLLDebug, { Loggit }
{$ENDIF}
   Messages, { WM_MENUCHAR, etc. }
   SysUtils, { StrLCopy, Format, etc. }
   Menus,    { TMenuItem }
   DUTUtilsSys; { IsAccel, GetHighlightedMenuItem }

//
const
{ QueryContextMenu uFlags }
  CMF_NORMAL             = $00000000;
//  CMF_DEFAULTONLY        = $00000001;
//  CMF_VERBSONLY          = $00000002;
//  CMF_EXPLORE            = $00000004;
//  CMF_NOVERBS            = $00000008;
//  CMF_CANRENAME          = $00000010;
//  CMF_NODEFAULT          = $00000020;
//  CMF_INCLUDESTATIC      = $00000040;
//  CMF_RESERVED           = $FFFF0000;      { View specific }

{ GetCommandString uFlags }
  GCS_VERBA            = $00000000;     { canonical verb }
  GCS_HELPTEXTA        = $00000001;     { help text (for status bar) }
  GCS_VALIDATEA        = $00000002;     { validate command exists }
//  GCS_VERBW            = $00000004;     { canonical verb (unicode) }
//  GCS_HELPTEXTW        = $00000005;     { help text (unicode version) }
//  GCS_VALIDATEW        = $00000006;     { validate command exists (unicode) }
//  GCS_UNICODE          = $00000004;     { for bit testing - Unicode string }

  GCS_VERB            = GCS_VERBA;
  GCS_HELPTEXT        = GCS_HELPTEXTA;
  GCS_VALIDATE        = GCS_VALIDATEA;

//  CMDSTR_NEWFOLDERA       = 'NewFolder';
//  CMDSTR_VIEWLISTA        = 'ViewList';
//  CMDSTR_VIEWDETAILSA     = 'ViewDetails';
//  CMDSTR_NEWFOLDERW       = 'NewFolder'; // !!! make WideString() ?
//  CMDSTR_VIEWLISTW        = 'ViewList';
//  CMDSTR_VIEWDETAILSW     = 'ViewDetails';

//  CMDSTR_NEWFOLDER        = CMDSTR_NEWFOLDERA;
//  CMDSTR_VIEWLIST         = CMDSTR_VIEWLISTA;
//  CMDSTR_VIEWDETAILS      = CMDSTR_VIEWDETAILSA;

//  CMDID_REFRESHDESKTOP    = $A065; //The shell's "Refresh Desktop" menu ID

//  CMIC_MASK_HOTKEY            = SEE_MASK_HOTKEY;
//  CMIC_MASK_ICON              = SEE_MASK_ICON;
//  CMIC_MASK_FLAG_NO_UI        = SEE_MASK_FLAG_NO_UI;
//  CMIC_MASK_UNICODE           = SEE_MASK_UNICODE;
//  CMIC_MASK_NO_CONSOLE        = SEE_MASK_NO_CONSOLE;
////  CMIC_MASK_HASLINKNAME       = SEE_MASK_HASLINKNAME; - not defined in shellapi
////  CMIC_MASK_FLAG_SEP_VDM      = SEE_MASK_FLAG_SEPVDM; - not defined in shellapi
////  CMIC_MASK_HASTITLE          = SEE_MASK_HASTITLE; - not defined in shellapi
//  CMIC_MASK_ASYNCOK           = SEE_MASK_ASYNCOK;
//

{ ########## Supporting types ########## }

type
   TIntegerObject = class(TObject)
   {
   TIntegerObject is a TObject descendant that holds an integer.
   It is a wrapper so we can use TStringList's Objects property.
   (Because simple typecasting won't work)
   }
   private
   { Unavailable }
      FInteger : Integer;
   public
   { Available at run-time }
      constructor Create(AInteger : Integer); virtual;
      property Value : Integer
         read FInteger
         write FInteger;
   end; { TIntegerObject }

   TNotifyObject = class(TObject)
   {
   TNotifyObject is a TObject descendant that holds an event handler.
   It is a wrapper so we can use TStringList's Objects property.
   (Because simple typecasting won't work)
   }
   private
   { Unavailable }
      FHandler : TNotifyEvent;
   public
   { Available at run-time }
      constructor Create(OnEvent : TNotifyEvent); virtual;
      property OnEvent : TNotifyEvent
         read FHandler
         write FHandler;
   end; { TNotifyObject }

{ ########## TIntList ########## }

function TIntList.Get(Index : Integer) : Integer;
begin
   Result := Integer(inherited Get(Index));
end; { TIntList.Get }

procedure TIntList.Put(Index, Item : Integer);
begin
   inherited Put(Index, Pointer(Item));
end; { TIntList.Put }

function TIntList.Add(Item : Integer) : Integer;
begin
   Result := inherited Add(Pointer(Item));
end; { TIntList.Add }

function TIntList.IndexOf(Item : Integer): Integer;
begin
   Result := inherited IndexOf(Pointer(Item));
end; { TIntList.IndexOf }

procedure TIntList.Insert(Index, Item : Integer);
begin
   inherited Insert(Index, Pointer(Item));
end; { TIntList.Insert }

function TIntList.Remove(Item : Integer) : Integer;
begin
   Result := inherited Remove(Pointer(Item));
end; { TIntList.Remove }

{ ########## TIntegerObject ########## }

constructor TIntegerObject.Create(AInteger : Integer);
begin
   inherited Create;
   FInteger := AInteger;
end; { constructor }

{ ########## TNotifyObject ########## }

constructor TNotifyObject.Create(OnEvent : TNotifyEvent);
begin
   inherited Create;
   FHandler := OnEvent;
end; { constructor }

{ ########## TContextMenu ########## }

//Protected stuff

constructor TContextMenu.Create;
begin
   inherited Create;
   FIDOffset := 0;
   FRootID := 0;

   FChecked := TDUTBits.Create;
   FGrayed := TDUTBits.Create;
   FHasSeparator := TDUTBits.Create;
   FVisible := TDUTBits.Create;

   FMenuBitmaps := TIntList.Create;
   FCheckedBitmaps := TIntList.Create;
   FUncheckedBitmaps := TIntList.Create;
   FOwnerDrawnData := TIntList.Create;

   FCaptions := TStringList.Create;
   FHints := TStringList.Create;
   FVerbs := TStringList.Create;
end; { constructor }

destructor TContextMenu.Destroy;
var
   LoopCount,
   Index : Integer;
begin
   LoopCount := Count - 1;

   for Index := 0 to LoopCount do begin
      //HBITMAPs
      //DeleteObject will fail if the hBitmap is Nil, but that's just fine with us
      DeleteObject(FMenuBitmaps[Index]);
      DeleteObject(FCheckedBitmaps[Index]);
      DeleteObject(FUncheckedBitmaps[Index]);
      //Submenus
      FCaptions.Objects[Index].Free;
      //ID
      FHints.Objects[Index].Free;
      //OnClick
      FVerbs.Objects[Index].Free;
   end; { for }

   FChecked.Free;
   FGrayed.Free;
   FHasSeparator.Free;
   FVisible.Free;

   FMenuBitmaps.Free;
   FCheckedBitmaps.Free;
   FUncheckedBitmaps.Free;
   FOwnerDrawnData.Free;

   FCaptions.Free;
   FHints.Free;
   FVerbs.Free;
   inherited Destroy;
end; { destructor }

function TContextMenu.GetCaption(Index : Integer) : String;
begin
   Result := FCaptions[Index];
end; { TContextMenu.GetCaption }

function TContextMenu.GetCaptionByID(CommandID : Integer) : String;
begin
   Result := ContextMenuOfID(CommandID).Captions[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetCaptionByID }

function TContextMenu.GetChecked(Index : Integer) : Boolean;
begin
   Result := FChecked.Bits[Index];
end; { TContextMenu.GetChecked }

function TContextMenu.GetCheckedByID(CommandID : Integer) : Boolean;
begin
   Result := ContextMenuOfID(CommandID).Checked[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetCheckedByID }

function TContextMenu.GetCheckedBitmap(Index : Integer) : HBITMAP;
begin
   Result := FCheckedBitmaps[Index];
end; { TContextMenu.GetCheckedBitmap }

function TContextMenu.GetCheckedBitmapByID(CommandID : Integer) : HBITMAP;
begin
   Result := ContextMenuOfID(CommandID).CheckedBitmaps[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetCheckedBitmapByID }

function TContextMenu.GetCount : Integer;
begin
   Result := FVerbs.Count;
end; { TContextMenu.GetCount }

function TContextMenu.GetGrayed(Index : Integer) : Boolean;
begin
   Result := FGrayed.Bits[Index];
end; { TContextMenu.GetGrayed }

function TContextMenu.GetGrayedByID(CommandID : Integer) : Boolean;
begin
   Result := ContextMenuOfID(CommandID).Grayed[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetGrayedByID }

function TContextMenu.GetHasCheckedBitmap(Index : Integer) : Boolean;
begin
   Result := (FCheckedBitmaps[Index] <> 0);
end; { TContextMenu.GetHasCheckedBitmap }

function TContextMenu.GetHasCheckedBitmapByID(CommandID : Integer) : Boolean;
begin
   Result := ContextMenuOfID(CommandID).HasCheckedBitmap[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetHasCheckedBitmapByID }

function TContextMenu.GetHasMenuBitmap(Index : Integer) : Boolean;
begin
   Result := (FMenuBitmaps[Index] <> 0);
end; { TContextMenu.GetHasMenuBitmap }

function TContextMenu.GetHasMenuBitmapByID(CommandID : Integer) : Boolean;
begin
   Result := ContextMenuOfID(CommandID).HasMenuBitmap[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetHasMenuBitmapByID }

function TContextMenu.GetHasSeparator(Index : Integer) : Boolean;
begin
   Result := FHasSeparator.Bits[Index];
end; { TContextMenu.GetHasSeparator }

function TContextMenu.GetHasSeparatorByID(CommandID : Integer) : Boolean;
begin
   Result := ContextMenuOfID(CommandID).HasSeparator[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetHasSeparatorByID }

function TContextMenu.GetHasSubmenu(Index : Integer) : Boolean;
begin
   Result := (FCaptions.Objects[Index] <> Nil);
end; { TContextMenu.GetHasSubmenu }

function TContextMenu.GetHasSubmenuByID(CommandID : Integer) : Boolean;
begin
   Result := ContextMenuOfID(CommandID).HasSubmenu[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetHasSubmenuByID }

function TContextMenu.GetHasUncheckedBitmap(Index : Integer) : Boolean;
begin
   Result := (FUncheckedBitmaps[Index] <> 0);
end; { TContextMenu.GetHasUncheckedBitmap }

function TContextMenu.GetHasUncheckedBitmapByID(CommandID : Integer) : Boolean;
begin
   Result := ContextMenuOfID(CommandID).HasUncheckedBitmap[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetHasUncheckedBitmapByID }

function TContextMenu.GetHint(Index : Integer) : String;
begin
   Result := FHints[Index];
end; { TContextMenu.GetHint }

function TContextMenu.GetHintByID(CommandID : Integer) : String;
begin
   Result := ContextMenuOfID(CommandID).Hints[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetHintByID }

function TContextMenu.GetID(Index : Integer) : Integer;
begin
   Result := (FHints.Objects[Index] as TIntegerObject).Value;
end; { TContextMenu.GetID }

function TContextMenu.GetIDCount : Integer;
{
   Returns 1 + the last ID used in the menu and its submenus
}
var
   Index : Integer;
begin
   Result := 0;
   //A RootID of -1 means the owning TContextMenu entry is invisible
   if RootID <> -1 then begin
      //Look for last ID used
      for Index := (Count - 1) downto 0 do begin
         //An ID of -1 means the entry is invisible
         if ID[Index] <> -1 then begin
            if HasSubmenu[Index] then begin
               Result := Submenu[Index].RootID + Submenu[Index].IDCount;
            end else begin
               Result := 1 + ID[Index];
            end; { else }
            break;
         end; { then }
      end; { for }
   end; { else }
end; { TContextMenu.GetIDCount }

function TContextMenu.GetIsOwnerDrawn(Index : Integer) : Boolean;
begin
   Result := (FOwnerDrawnData[Index] <> 0);
end; { TContextMenu.GetIsOwnerDrawn }

function TContextMenu.GetIsOwnerDrawnByID(CommandID : Integer) : Boolean;
begin
   Result := ContextMenuOfID(CommandID).IsOwnerDrawn[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetIsOwnerDrawnByID }

function TContextMenu.GetMenuBitmap(Index : Integer) : HBITMAP;
begin
   Result := FMenuBitmaps[Index];
end; { TContextMenu.GetMenuBitmap }

function TContextMenu.GetMenuBitmapByID(CommandID : Integer) : HBITMAP;
begin
   Result := ContextMenuOfID(CommandID).MenuBitmaps[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetMenuBitmapByID }

function TContextMenu.GetOnClick(Index : Integer) : TNotifyEvent;
begin
   Result := (FVerbs.Objects[Index] as TNotifyObject).OnEvent;
end; { TContextMenu.GetOnClick }

function TContextMenu.GetOnClickByID(CommandID : Integer) : TNotifyEvent;
begin
   Result := ContextMenuOfID(CommandID).OnClick[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetOnClickByID }

function TContextMenu.GetOwnerDrawnData(Index : Integer) : Integer;
begin
   Result := FOwnerDrawnData[Index];
end; { TContextMenu.GetOwnerDrawnData }

function TContextMenu.GetOwnerDrawnDataByID(CommandID : Integer) : Integer;
begin
   Result := ContextMenuOfID(CommandID).OwnerDrawnData[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetOwnerDrawnDataByID }

function TContextMenu.GetSubmenu(Index : Integer) : TContextMenu;
begin
   Result := TContextMenu(FCaptions.Objects[Index]);
end; { TContextMenu.GetSubmenu }

function TContextMenu.GetSubmenuByID(CommandID : Integer) : TContextMenu;
begin
   Result := ContextMenuOfID(CommandID).Submenu[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetSubmenuByID }

function TContextMenu.GetTotalCount : Integer;
{
   Returns the Count of the TContextItem plus that of all of its submenus,
   even the Invisible ones.
}
var
   Index : Integer;
begin
   Result := Count;
   for Index := 0 to (Count - 1) do begin
      if HasSubmenu[Index] then Result := Result + Submenu[Index].TotalCount;
   end; { for }
end; { TContextMenu.GetTotalCount }

function TContextMenu.GetUncheckedBitmap(Index : Integer) : HBITMAP;
begin
   Result := FUncheckedBitmaps[Index];
end; { TContextMenu.GetUncheckedBitmap }

function TContextMenu.GetUncheckedBitmapByID(CommandID : Integer) : HBITMAP;
begin
   Result := ContextMenuOfID(CommandID).UncheckedBitmaps[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetUncheckedBitmapByID }

function TContextMenu.GetVerb(Index : Integer) : String;
begin
   Result := FVerbs[Index];
end; { TContextMenu.GetVerb }

function TContextMenu.GetVerbByID(CommandID : Integer) : String;
begin
   Result := ContextMenuOfID(CommandID).Verbs[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetVerbByID }

function TContextMenu.GetVisible(Index : Integer) : Boolean;
begin
   Result := FVisible.Bits[Index];
end; { TContextMenu.GetVisible }

function TContextMenu.GetVisibleByID(CommandID : Integer) : Boolean;
{
   Since Invisible menu entries don't have IDs, this'll either return True or raise an exception.
}
begin
   Result := ContextMenuOfID(CommandID).Visible[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)];
end; { TContextMenu.GetVisibleByID }

procedure TContextMenu.ResetIDs;
{
   Sets the IDs of the menu and of all its submenus,
   sets the RootIDs of the submenus.
   The menu entries have IDs starting at zero,
   a submenu header's ID is immediately followed
   by the IDs of the submenu entries.
   Example:
   (RootID has some value)
   mnuA (ID 0)
   mnuB (ID 1)
      (RootID 2)
      mnuBA (ID 0) --the ID's at mnuX level are (mnuXX ID + mnuXX RootID)
      mnuBB (ID 1)
   mnuC (ID 4)

   An invisible entry has an ID of -1 (for "none"); any owned
   submenus also use no IDs; this is indicated by a RootID of -1.
}
var
   CurrentID,
   Index : Integer;
begin
   CurrentID := 0;
   for Index := 0 to (Count - 1) do begin
      if Visible[Index] then begin
         SetID(Index, CurrentID);
         Inc(CurrentID);
         if HasSubmenu[Index] then begin
            Submenu[Index].RootID := CurrentID;
            Submenu[Index].ResetIDs;
            CurrentID := CurrentID + Submenu[Index].IDCount;
         end; { then }
      end else begin
         SetID(Index, -1);
         if HasSubmenu[Index] then Submenu[Index].ZapIDs;
      end; { else }
   end; { for }
end; { TContextMenu.ResetIDs }

procedure TContextMenu.SetCaption(Index : Integer; Value : String);
begin
   FCaptions[Index] := Value;
end; { TContextMenu.SetCaption }

procedure TContextMenu.SetCaptionByID(CommandID : Integer; Value : String);
begin
   ContextMenuOfID(CommandID).Captions[ContextMenuOfID(CommandID).IndexOfID(CommandID)] := Value;
end; { TContextMenu.SetCaptionByID }

procedure TContextMenu.SetChecked(Index : Integer; Value : Boolean);
begin
   FChecked.Bits[Index] := Value;
end; { TContextMenu.SetChecked }

procedure TContextMenu.SetCheckedByID(CommandID : Integer; Value : Boolean);
begin
   ContextMenuOfID(CommandID).Checked[ContextMenuOfID(CommandID).IndexOfID(CommandID)] := Value;
end; { TContextMenu.SetCheckedByID }

procedure TContextMenu.SetCheckedBitmap(Index : Integer; Value : HBITMAP);
begin
   //Free any previous HBITMAP
   DeleteObject(FCheckedBitmaps[Index]);
   FCheckedBitmaps[Index] := Value;
end; { TContextMenu.SetCheckedBitmap }

procedure TContextMenu.SetCheckedBitmapByID(CommandID : Integer; Value : HBITMAP);
begin
   ContextMenuOfID(CommandID).CheckedBitmaps[ContextMenuOfID(CommandID).IndexOfID(CommandID)] := Value;
end; { TContextMenu.SetCheckedBitmapsByID }

procedure TContextMenu.SetGrayed(Index : Integer; Value : Boolean);
begin
   FGrayed.Bits[Index] := Value;
end; { TContextMenu.SetGrayed }

procedure TContextMenu.SetGrayedByID(CommandID : Integer; Value : Boolean);
begin
   ContextMenuOfID(CommandID).Grayed[ContextMenuOfID(CommandID).IndexOfID(CommandID)] := Value;
end; { TContextMenu.SetGrayedByID }

procedure TContextMenu.SetHasSeparator(Index : Integer; Value : Boolean);
begin
   FHasSeparator.Bits[Index] := Value;
end; { TContextMenu.SetHasSeparator }

procedure TContextMenu.SetHasSeparatorByID(CommandID : Integer; Value : Boolean);
begin
   ContextMenuOfID(CommandID).HasSeparator[ContextMenuOfID(CommandID).IndexOfID(CommandID)] := Value;
end; { TContextMenu.SetHasSeparatorByID }

procedure TContextMenu.SetHint(Index : Integer; Value : String);
begin
   FHints[Index] := Value;
end; { TContextMenu.SetHint }

procedure TContextMenu.SetHintByID(CommandID : Integer; Value : String);
begin
   ContextMenuOfID(CommandID).Hints[ContextMenuOfID(CommandID).IndexOfID(CommandID)] := Value;
end; { TContextMenu.SetHintByID }

procedure TContextMenu.SetID(Index : Integer; Value : Integer);
begin
   (FHints.Objects[Index] as TIntegerObject).Value := Value;
end; { TContextMenu.SetID }

procedure TContextMenu.SetMenuBitmap(Index : Integer; Value : HBITMAP);
begin
   //Free any previous HBITMAP
   DeleteObject(FMenuBitmaps[Index]);
   FMenuBitmaps[Index] := Value;
end; { TContextMenu.SetMenuBitmap }

procedure TContextMenu.SetMenuBitmapByID(CommandID : Integer; Value : HBITMAP);
begin
   ContextMenuOfID(CommandID).MenuBitmaps[ContextMenuOfID(CommandID).IndexOfID(CommandID)] := Value;
end; { TContextMenu.SetMenuBitmapsByID }

procedure TContextMenu.SetOnClick(Index : Integer; Value : TNotifyEvent);
begin
   //Free any previous TNotifyObject
   FVerbs.Objects[Index].Free;
   FVerbs.Objects[Index] := TNotifyObject.Create(Value);
end; { TContextMenu.SetOnClick }

procedure TContextMenu.SetOnClickByID(CommandID : Integer; Value : TNotifyEvent);
begin
   ContextMenuOfID(CommandID).OnClick[ContextMenuOfID(CommandID).IndexOfID(CommandID)] := Value;
end; { TContextMenu.SetOnClickByID }

procedure TContextMenu.SetOwnerDrawnData(Index, Value : Integer);
begin
   FOwnerDrawnData[Index] := Value;
end; { TContextMenu.SetOwnerDrawnData }

procedure TContextMenu.SetOwnerDrawnDataByID(CommandID, Value : Integer);
begin
   ContextMenuOfID(CommandID).OwnerDrawnData[ContextMenuOfID(CommandID).IndexOfID(CommandID - ContextMenuOfID(CommandID).RootID)] := Value;
end; { TContextMenu.SetOwnerDrawnDataByID }

procedure TContextMenu.SetSubmenu(Index : Integer; Value : TContextMenu);
begin
   //Free any previous TContextMenu
   FCaptions.Objects[Index].Free;
   FCaptions.Objects[Index] := Value;
   ResetIDs;
end; { TContextMenu.SetSubmenu }

procedure TContextMenu.SetSubmenuByID(CommandID : Integer; Value : TContextMenu);
begin
   ContextMenuOfID(CommandID).Submenu[ContextMenuOfID(CommandID).IndexOfID(CommandID)] := Value;
end; { TContextMenu.SetSubmenuByID }

procedure TContextMenu.SetUncheckedBitmap(Index : Integer; Value : HBITMAP);
begin
   //Free any previous HBITMAP
   DeleteObject(FUncheckedBitmaps[Index]);
   FUncheckedBitmaps[Index] := Value;
end; { TContextMenu.SetUncheckedBitmap }

procedure TContextMenu.SetUncheckedBitmapByID(CommandID : Integer; Value : HBITMAP);
begin
   ContextMenuOfID(CommandID).UncheckedBitmaps[ContextMenuOfID(CommandID).IndexOfID(CommandID)] := Value;
end; { TContextMenu.SetUncheckedBitmapsByID }

procedure TContextMenu.SetVerb(Index : Integer; Value : String);
begin
   FVerbs[Index] := Value;
end; { TContextMenu.SetVerb }

procedure TContextMenu.SetVerbByID(CommandID : Integer; Value : String);
begin
   ContextMenuOfID(CommandID).Verbs[ContextMenuOfID(CommandID).IndexOfID(CommandID)] := Value;
end; { TContextMenu.SetVerbByID }

procedure TContextMenu.SetVisible(Index : Integer; Value : Boolean);
begin
   FVisible.Bits[Index] := Value;
   ResetIDs;
end; { TContextMenu.SetVisible }

procedure TContextMenu.SetVisibleByID(CommandID : Integer; Value : Boolean);
begin
   ContextMenuOfID(CommandID).Visible[ContextMenuOfID(CommandID).IndexOfID(CommandID)] := Value;
end; { TContextMenu.SetVisibleByID }

procedure TContextMenu.ZapIDs;
{
   Sets the IDs of the menu and of all its submenus to -1,
   sets the RootIDs of the submenus to -1.
}
var
   Index : Integer;
begin
   RootID := -1;
   for Index := 0 to (Count - 1) do begin
      SetID(Index, -1);
      if HasSubmenu[Index] then Submenu[Index].ZapIDs;
   end; { for }
end; { TContextMenu.ZapIDs }

//Public stuff

function TContextMenu.Add(const Verb, Caption, Hint : string; const OnClick : TNotifyEvent; const Checked, Grayed, Separator, Visible : Boolean) : Integer;
{
   Adds a menu entry to the TContextMenu.
   The input parametres should be pretty much explanatory.
   Verb can also be used as a Name; it ought to exclude any embedded blanks.
   Separator, when True, means a separator follows this entry
   (separators do not occupy menu slots in TContextMenu).
}
begin
   //ID --Must do this first so GetIDCount does not get confused.
   if Visible then begin
      FHints.AddObject(Hint, TIntegerObject.Create(IDCount));
   end else begin
      FHints.AddObject(Hint, TIntegerObject.Create(-1));
   end; { else }

   FChecked.Add(Checked);
   FGrayed.Add(Grayed);
   FHasSeparator.Add(Separator);
   FVisible.Add(Visible);

   FMenuBitmaps.Add(0);
   FCheckedBitmaps.Add(0);
   FUncheckedBitmaps.Add(0);
   FOwnerDrawnData.Add(0);

   //No submenu to add
   FCaptions.Add(Caption);
   //OnClick
   Result := FVerbs.AddObject(Verb, TNotifyObject.Create(OnClick));
end; { TContextMenu.Add }

procedure TContextMenu.Assign(Source : TPersistent);
{
   Supports assignment of a TMenuItem to TContextMenu (such as TMainMenu.Items or TPopupMenu.Items).
   Note that IDs cannot be assigned and that TMenuItems don't support CheckedBitmaps or MenuBitmaps.
}
var
   Index    : Integer;
   ASubmenu : TContextMenu;
begin
   if Source is TMenuItem then begin
      Clear;
      for Index := 0 to ((Source as TMenuItem).Count - 1) do begin
         if (Source as TMenuItem).Items[Index].Caption = '-' then begin
            {
            TMenuItem stores separators as entries where the caption is a hyphen;
            we store them as flags in the previous entry.
            We must ignore any separator that appears as the first entry;
            we'll also collapse multiple separator sequences into single separators.
            }
            if Index > 0 then HasSeparator[Count - 1] := True;
         end else begin
            Add((Source as TMenuItem).Items[Index].Name,
                (Source as TMenuItem).Items[Index].Caption,
                (Source as TMenuItem).Items[Index].Hint,
                (Source as TMenuItem).Items[Index].OnClick,
                (Source as TMenuItem).Items[Index].Checked,
                not (Source as TMenuItem).Items[Index].Enabled, { Grayed }
                False, { Separator }
                (Source as TMenuItem).Items[Index].Visible);
            if (Source as TMenuItem).Items[Index].Count > 0 then begin
               { There is a submenu }
               ASubmenu := TContextMenu.Create;
               { Recursively assign the submenu }
               ASubmenu.Assign((Source as TMenuItem).Items[Index]);
               { Attach it }
               Submenu[Count - 1] := ASubmenu;
            end; { then }
         end; { else }
      end; { for }
   end else begin
      inherited;
   end; { else }
end; { TContextMenu.Assign }

procedure TContextMenu.AssignTo(Dest: TPersistent);
{
   Supports assignment to a TMenuItem from TContextMenu.
   Note that CommandIDs are beyond our control here and that
   CheckedBitmaps and MenuBitmaps cannot be exported.
}
var
   Index    : Integer;
   ASubmenu : TMenuItem;
begin
   if Dest is TMenuItem then begin
      //Clear the destination
      while (Dest as TMenuItem).Count > 0 do (Dest as TMenuItem).Delete(0);
      //Rebuild it
      for Index := 0 to (Count - 1) do begin
         (Dest as TMenuItem).Add(TMenuItem.Create(TComponent(Dest)));
         (Dest as TMenuItem)[(Dest as TMenuItem).Count - 1].Name := Verbs[Index];
         (Dest as TMenuItem)[(Dest as TMenuItem).Count - 1].Caption := Captions[Index];
         (Dest as TMenuItem)[(Dest as TMenuItem).Count - 1].Checked := Checked[Index];
         (Dest as TMenuItem)[(Dest as TMenuItem).Count - 1].Enabled := not Grayed[Index];
         { GroupIndex? }
         { HelpContext? }
         (Dest as TMenuItem)[(Dest as TMenuItem).Count - 1].Hint := Hints[Index];
         { RadioItem? }
         { Shortcut? }
         (Dest as TMenuItem)[(Dest as TMenuItem).Count - 1].Visible := Visible[Index];
         (Dest as TMenuItem)[(Dest as TMenuItem).Count - 1].OnClick := OnClick[Index];
         if HasSubmenu[Index] then begin
            { There is a submenu }
            ASubmenu := TMenuItem.Create((Dest as TMenuItem)[(Dest as TMenuItem).Count - 1]);
            { Recursively assign to the submenu }
            Submenu[Index].AssignTo(ASubmenu);
         end; { then }
         if HasSeparator[Index] then begin
            {
            TMenuItem stores separators as entries where the caption is a hyphen;
            we store them as flags in the previous entry.
            }
            (Dest as TMenuItem).Add(TMenuItem.Create(TComponent(Dest)));
            (Dest as TMenuItem)[(Dest as TMenuItem).Count - 1].Caption := '-';
         end; { then }
      end; { for }
   end else begin
      inherited;
   end; { else }
end; { TContextMenu.AssignTo }

procedure TContextMenu.Clear;
{
   Empties the TContextMenu.  Any submenus are disposed of.
}
var
   LoopCount,
   Index : Integer;
begin
   LoopCount := Count - 1;

   for Index := 0 to LoopCount do begin
      //HBITMAPS
      //DeleteObject will fail if the hBitmap is Nil, but that's just fine with us
      DeleteObject(FMenuBitmaps[Index]);
      DeleteObject(FCheckedBitmaps[Index]);
      DeleteObject(FUncheckedBitmaps[Index]);
      //Submenus
      FCaptions.Objects[Index].Free;
      //ID
      FHints.Objects[Index].Free;
      //OnClick
      FVerbs.Objects[Index].Free;
   end; { for }

   FChecked.Clear;
   FGrayed.Clear;
   FHasSeparator.Clear;
   FVisible.Clear;

   FMenuBitmaps.Clear;
   FCheckedBitmaps.Clear;
   FUncheckedBitmaps.Clear;
   FOwnerDrawnData.Clear;

   FCaptions.Clear;
   FHints.Clear;
   FVerbs.Clear;
end; { TContextMenu.Clear }

function TContextMenu.ContextMenuOfID(const CommandID : Integer) : TContextMenu;
{
   Returns the TContextMenu that contains CommandID.
   This can be Self, one of its submenus, or Nil.
}
var
   Index : Integer;
begin
   Result := Nil;
   if IndexOfID(CommandID) >= 0 then begin
      Result := Self;
   end else begin
      { Check submenus }
      for Index := 0 to (Count - 1) do begin
         if HasSubmenu[Index] then begin
            Result := Submenu[Index].ContextMenuOfID(CommandID - Submenu[Index].RootID);
            if Assigned(Result) then break;
         end; { then }
      end; { for }
   end; { else }
end; { TContextMenu.ContextMenuOfID }

procedure TContextMenu.Delete(Index : Integer);
{
   Deletes the specified entry.
}
begin
   //HBITMAPS
   DeleteObject(FMenuBitmaps[Index]);
   DeleteObject(FCheckedBitmaps[Index]);
   DeleteObject(FUncheckedBitmaps[Index]);
   //Submenus
   FCaptions.Objects[Index].Free;
   //IDs
   FHints.Objects[Index].Free;
   //OnClick
   FVerbs.Objects[Index].Free;

   FChecked.Delete(Index);
   FGrayed.Delete(Index);
   FHasSeparator.Delete(Index);
   FVisible.Delete(Index);

   FMenuBitmaps.Delete(Index);
   FCheckedBitmaps.Delete(Index);
   FUncheckedBitmaps.Delete(Index);
   FOwnerDrawnData.Delete(Index);

   FCaptions.Delete(Index);
   FHints.Delete(Index);
   FVerbs.Delete(Index);

   ResetIDs;
end; { TContextMenu.Delete }

procedure TContextMenu.Exchange(Index1, Index2 : Integer);
{
   Exchanges the two specified entries.  Submenus, if any, are swapped too.
}
var
   CommandID : Integer;
begin
   FChecked.Exchange(Index1, Index2);
   FGrayed.Exchange(Index1, Index2);
   FHasSeparator.Exchange(Index1, Index2);
   FVisible.Exchange(Index1, Index2);

   FMenuBitmaps.Exchange(Index1, Index2);
   FCheckedBitmaps.Exchange(Index1, Index2);
   FUncheckedBitmaps.Exchange(Index1, Index2);
   FOwnerDrawnData.Exchange(Index1, Index2);

   //Submenus will be swapped too
   FCaptions.Exchange(Index1, Index2);
   FHints.Exchange(Index1, Index2);
   //We do NOT want the IDs to swap, so we must un-swap them
   CommandID := ID[Index1];
   SetID(Index1, ID[Index2]);
   SetID(Index2, CommandID);
   //OnClick will be swapped too
   FVerbs.Exchange(Index1, Index2);
end; { TContextMenu.Exchange }

function TContextMenu.IDOfString(const AString : String; const StringType : TContextMenuStringType) : Integer;
{
   Returns the ID of the first item in TContextMenu and its submenus
   that contains AString in its StringType property array.
   Returns -1 if no such match exists.
   Case-sensitive.
}
var
   Index : Integer;
begin
   Result := -1;
   case StringType of
   cmstCaption : begin
      if IndexOfCaption(AString) >= 0 then begin
         Result := ID[IndexOfCaption(AString)];
         Exit;
      end; { then }
      end; { cmstCaption }
   cmstHint : begin
      if IndexOfHint(AString) >= 0 then begin
         Result := ID[IndexOfHint(AString)];
         Exit;
      end; { then }
      end; { cmstHint }
   cmstVerb : begin
      if IndexOfVerb(AString) >= 0 then begin
         Result := ID[IndexOfVerb(AString)];
         Exit;
      end; { then }
      end; { cmstVerb }
   end; { case }
   { Check submenus }
   for Index := 0 to (Count - 1) do begin
      if HasSubmenu[Index] then begin
         Result := Submenu[Index].IDOfString(AString, StringType) + Submenu[Index].RootID;
         if Result >= Submenu[Index].RootID then break;
      end; { then }
   end; { for }
end; { TContextMenu.IDOfString }

function TContextMenu.IndexOfCaption(const Caption : string): Integer;
{
   This'll return -1 if the Caption sought is in a submenu.
}
begin
   Result := FCaptions.IndexOf(Caption);
end; { TContextMenu.IndexOfCaption }

function TContextMenu.IndexOfHint(const Hint : string): Integer;
{
   This'll return -1 if the Hint sought is in a submenu.
}
begin
   Result := FHints.IndexOf(Hint);
end; { TContextMenu.IndexOfHint }

function TContextMenu.IndexOfID(const CommandID : Integer) : Integer;
{
   This'll return -1 if the ID sought is in a submenu.
}
var
   Index : Integer;
begin
   Result := -1;
   for Index := 0 to (Count - 1) do begin
      if ID[Index] = CommandID then begin
         Result := Index;
         break;
      end; { then }
   end; { for }
end; { TContextMenu.IndexOfID }

function TContextMenu.IndexOfVerb(const Verb : string): Integer;
{
   This'll return -1 if the Verb sought is in a submenu.
}
begin
   Result := FVerbs.IndexOf(Verb);
end; { TContextMenu.IndexOfVerb }

procedure TContextMenu.Insert(Index : Integer; const Verb, Caption, Hint : string; const OnClick : TNotifyEvent; const Checked, Grayed, Separator, Visible : Boolean);
{
   Inserts an entry at the specified Index.
   That is, entries at Index and higher are pushed one up.
}
begin
   FChecked.Insert(Index, Checked);
   FGrayed.Insert(Index, Grayed);
   FHasSeparator.Insert(Index, Separator);
   FVisible.Insert(Index, Visible);

   FMenuBitmaps.Insert(Index, 0);
   FCheckedBitmaps.Insert(Index, 0);
   FUncheckedBitmaps.Insert(Index, 0);
   FOwnerDrawnData.Insert(Index, 0);

   //Submenus remains Nil
   FCaptions.Insert(Index, Caption);
   //ID --assign -1 until ResetIDs figures it out
   FHints.InsertObject(Index, Hint, TIntegerObject.Create(-1));
   //OnClick
   FVerbs.InsertObject(Index, Verb, TNotifyObject.Create(OnClick));

   ResetIDs;
end; { TContextMenu.Insert }

function TContextMenu.InsertInHMenu(var Menu : HMENU; var IndexOffset : Integer) : Boolean;
{
   Inserts the TContextMenu into the specified Windows menu at the specified IndexOffset.
   The CommandIDs are based on the TContextMenu's FIDOffset + ID[]
   Menu              Menu handle to use
   IndexOffset       Offset into the menu where to begin inserting items
   Returns False in case of catastrophic failure (couldn't create or insert a menu item).
}
   function SetMenuItemID(Menu : HMENU; IndexOffset, CommandID : Integer) : Boolean;
   var
      MenuItemInfo : TMenuItemInfo;
   begin
      Result := False;
      MenuItemInfo.cbSize := SizeOf(MenuItemInfo);
      MenuItemInfo.fMask := MIIM_ID;
      if not GetMenuItemInfo(Menu, IndexOffset, True {MF_BYPOSITION}, MenuItemInfo) then Exit;
      MenuItemInfo.wID := CommandID;
      Result := SetMenuItemInfo(Menu, IndexOffset, True {MF_BYPOSITION}, MenuItemInfo);
   end; { SetMenuItemID }

var
   CurrentID,
   SubIndexOffset,
   Index        : Integer;
   Flags        : UINT;
   hSubMenu     : HMENU;
   WasSeparator : Boolean; //Whether the last insertion was a separator
begin
{$IFDEF Debug}
Loggit('InsertInHMenu Begins HMenu = ' + IntToStr(Menu));
{$ENDIF}
   //Assume the worst
   Result := False;
   //Validate handle
   if not IsMenu(Menu) then Exit;
   hSubMenu := 0;
   CurrentID := IDOffset;
   WasSeparator := True;
   for Index := 0 to (Count - 1) do begin
      Flags := MF_BYPOSITION;
      //Skip invisible entries
      if Visible[Index] then begin
         if Grayed[Index] then Flags := Flags or MF_GRAYED;
         if Checked[Index] then Flags := Flags or MF_CHECKED;
         //Insert the menu item
         {
         I've tried inserting the menu, then attaching a
         submenu with ModifyMenu but that does not work.
         }
         //Submenu to attach?
         if HasSubmenu[Index] and (Submenu[Index].IDCount > 0) then begin
            Flags := Flags or MF_POPUP;
            //Create the submenu
            hSubMenu := CreatePopupMenu;
            SubIndexOffset := 0; //Submenu starts inserting at zero (its root)
            Submenu[Index].IDOffset := CurrentID + 1; //Must pass absolute ID offsets...
            if not Submenu[Index].InsertInHMenu(hSubMenu, SubIndexOffset) then break;
            if HasMenuBitmap[Index] then begin
               if not InsertMenu(Menu, IndexOffset, MF_BITMAP or Flags, hSubMenu, PChar(MenuBitmaps[Index])) then break;
            end else if IsOwnerDrawn[Index] then begin
{$IFDEF Debug}
Loggit('IsOwnerDrawn');
{$ENDIF}
               if not InsertMenu(Menu, IndexOffset, MF_OWNERDRAW or Flags, hSubMenu, PChar(OwnerDrawnData[Index])) then break;
            end else begin
               if not InsertMenu(Menu, IndexOffset, MF_STRING or Flags, hSubMenu, PChar(Captions[Index])) then break;
            end; { else }
            if not SetMenuItemID(Menu, IndexOffset, CurrentID) then break;
            hSubMenu := 0;
            CurrentID := CurrentID + Submenu[Index].IDCount;
         end else begin
            if HasMenuBitmap[Index] then begin
               if not InsertMenu(Menu, IndexOffset, MF_BITMAP or Flags, CurrentID, PChar(MenuBitmaps[Index])) then break;
            end else if IsOwnerDrawn[Index] then begin
{$IFDEF Debug}
Loggit('IsOwnerDrawn');
{$ENDIF}
               if not InsertMenu(Menu, IndexOffset, MF_OWNERDRAW or Flags, CurrentID, PChar(OwnerDrawnData[Index])) then break;
            end else begin
               if not InsertMenu(Menu, IndexOffset, MF_STRING or Flags, CurrentID, PChar(Captions[Index])) then break;
            end; { else }
         end; { else }
         //Add Checked/Unchecked bitmaps if so instructed
         if HasCheckedBitmap[Index] or HasUncheckedBitmap[Index] then begin
            //if SetMenuItemBitmaps fails, use default checkmark
            SetMenuItemBitmaps(Menu, IndexOffset, MF_BYPOSITION, UncheckedBitmaps[Index], CheckedBitmaps[Index]);
         end; { then }
         Inc(CurrentID);
         WasSeparator := False;
         Inc(IndexOffset);
      end; { then }
      //Add a separator if so instructed
      //An Invisible entry adds a separator only if the last previous insertion wasn't one
      if HasSeparator[Index] and
         (Visible[Index] or (not WasSeparator)) then begin
         if not InsertMenu(Menu, IndexOffset, MF_SEPARATOR or Flags, 0, '') then break;
         Inc(IndexOffset);
         WasSeparator := True;
      end; { then }
      //Success?
      Result := (Index = (Count - 1));
   end; { for }
{$IFDEF Debug}
Loggit('InsertInHMenu loop ends...');
{$ENDIF}
   //Did we succeed?
   if Result then begin
      //Check to be sure the last addition isn't a dangling separator
      if WasSeparator and (GetMenuItemCount(Menu) <= IndexOffset) then begin
         Dec(IndexOffset);
         Result := DeleteMenu(Menu, IndexOffset, MF_BYPOSITION);
      end; { then }
   end else begin
{$IFDEF Debug}
Loggit('...in failure');
{$ENDIF}
      //Failure: Clean up used resources
      DestroyMenu(Menu);
      if IsMenu(hSubMenu) then DestroyMenu(hSubMenu);
      Menu := 0;
      hSubMenu := 0;
   end; { else }
end; { TContextMenu.InsertInHMenu }

{ ########## TCOMObjContextMenu ########## }

function TCOMObjContextMenu.QueryContextMenu(Menu : HMENU; indexMenu, idCmdFirst, idCmdLast, uFlags : UINT) : HResult;
{
   QueryContextMenu is called by the shell right before displaying the context menu.

Menu
   Handle of the menu.
   The handler should specify this handle when calling the InsertMenu or InsertMenuItem function.
indexMenu
   Zero-based position at which to insert the first menu item.
idCmdFirst and idCmdLast
   Minimum and maximum values that the handler can specify for menu item identifiers.
   The actual identifier of each menu item should be idCmdFirst plus a menu identifier offset in the range zero through (idCmdLast - idCmdFirst).
uFlags
   Flag specifying zero or more of the following values:
CMF_DEFAULTONLY 1 The user is activating the default action, typically by double-clicking.
                  A context menu extension or drag and drop handler should not add any menu items if this value is specified (because the menu won't be displayed).
                  A name space extension should add only the default item (if any).
CMF_EXPLORE 	 4 Context menu handlers should ignore this value.
                  This value is specified if the context menu is for an object in the left pane of the explorer.
CMF_NORMAL	    0 Indicates normal operation.
                  A context menu extension, name space extension, or drag and drop handler can add any menu items.
CMF_VERBSONLY	 2 Context menu handlers should ignore this value.
                  This value is specified if the context menu is for a shortcut object.
   The remaining bits of the low-order word are reserved by the system:
CMF_NOVERBS     8      CMF_CANRENAME     10
CMF_NODEFAULT  20      CMF_INCLUDESTATIC 40
   The high-order word (CMF_RESERVED) may be used for context-specific communications.
}
var
   IndexOffset : Integer;
begin
{$IFDEF Debug}
Loggit('QueryContextMenu Menu = ' + IntToStr(Menu) + ' idCmdFirst = ' + IntToStr(idCmdFirst));
{$ENDIF}
   Result := S_OK; // MakeResult(SEVERITY_SUCCESS, FACILITY_NULL, 0);
   if CMF_NORMAL <> (CMF_NORMAL and uFlags) then Exit;

   FContextMenu.Free;
   FContextMenu := TContextMenu.Create;
   try
      Result := DoQuery;
   except
      Result := E_FAIL;
   end;
   if (Result <> S_OK) then Exit;
{
   An HRESULT is made up of a 1-bit severity flag, an 11-bit handle (implicit), a 4-bit facility code indicating status code (SCODE) group, and a 16-bit SCODE information code.
   QueryContextMenu expects (if the method is successful) the code member to contain the menu identifier offset of the next menu item to be added.
   That is to say, the idCmdFirst for the next handler.
}
   //Note that separators do not use any ID slots, so they're not counted
   //Neither are "invisible" items, since they weren't added to the menu
   IndexOffset := indexMenu;
   FContextMenu.IDOffset := idCmdFirst;
   if (FContextMenu.IDCount <= (1 + idCmdLast - idCmdFirst)) and
       FContextMenu.InsertInHMenu(Menu, IndexOffset) then begin
      Result := MakeResult(SEVERITY_SUCCESS, FACILITY_NULL, idCmdFirst + FContextMenu.IDCount);
   end else begin
      //Not enough slots available OR hMenu insertion failed
      Result := MakeResult(SEVERITY_ERROR, FACILITY_NULL, 0); //No idea how to pick error code...
   end; { else }
end; { IContextMenu.QueryContextMenu }

function TCOMObjContextMenu.GetCommandString(idCmd, uType : UINT; pwReserved : PUINT; pszName : LPSTR; cchMax : UINT) : HRESULT;
{
   GetCommandString is used to obtain help text for display in the Explorer's status bar
   and to obtain the string to pass to InvokeCommand.

idCmd
   Menu item identifier offset.
uType
   Flag specifying the information to retrieve.
   This parameter can be one of the following values:
GCS_HELPTEXT	Return the help text for the menu item.
GCS_VALIDATE	Validate that the menu item exists.
GCS_VERB	      Return the language-independent command name for the menu item.
pwReserved
   Reserved.
   Applications must specify NULL when calling this method, and handlers must ignore this parameter when called.
pszName and cchMax
   Address and size of the buffer that receives the returned null-terminated string.
}
begin
{$IFDEF Debug}
Loggit('GetCommandString Begins; idCmd = ' + IntToStr(idCmd) + ' uType = ' + IntToStr(uType));
{$ENDIF}
   try
      case uType of
      GCS_HELPTEXT {1} : begin
         //Return help string for menu item
         StrLCopy(pszName, PChar(FContextMenu.HintsByID[idCmd]), cchMax);
         end; { GCS_HELPTEXT }
      GCS_VALIDATE {2} : begin
         { No action? }
{$IFDEF Debug}
Loggit('Validate requested');
{$ENDIF}
         end; { GCS_VALIDATE }
      GCS_VERB {0} : begin
         //Return the shell action verb (that will be passed to InvokeCommand)
         StrLCopy(pszName, PChar(FContextMenu.VerbsByID[idCmd]), cchMax);
         end; { GCS_VERB }
      else
{$IFDEF Debug}
Loggit('Other uType requested = ' + IntToStr(uType));
{$ENDIF}
      end; { case }
      Result := NOERROR;
   except
      Result := E_INVALIDARG;
   end; { try }
{$IFDEF Debug}
Loggit('GetCommandString Ends');
{$ENDIF}
end; { IContextMenu.GetCommandString }

function TCOMObjContextMenu.InvokeCommand(var lpici : TCMInvokeCommandInfo) : HResult;
{
   InvokeCommand is called by the shell if the user selects one of the extension-added menu items.

lpici
   Pointer to a CMINVOKECOMMANDINFO structure containing information about the command.
   The fields are:
cbSize         sizeof(CMINVOKECOMMANDINFO)
fMask          Value specifying zero or more of the following flags:
   CMIC_MASK_HOTKEY	   Specifies that the dwHotKey member is valid.
   CMIC_MASK_ICON	      Specifies that the hIcon member is valid.
   CMIC_MASK_FLAG_NO_UI Prevents the system from displaying user interface elements (for example, error messages) while carrying out a command.
hWnd           Handle of the window that owned the context menu, such as the desktop, the explorer, or the tray.
               An extension might specify this handle as the owner window of any message boxes or dialog boxes it displays.
lpVerb         A 32-bit value containing zero in the high-order word and the menu-identifier offset of the command to carry out in the low-order word.
               The shell specifies this value when the user chooses a menu command.
               If the high-order word is not zero, this member is the address of a null-terminated string specifying the language-independent name of the command to carry out.
               This member is typically a string when a command is being activated by an application.
               The system provides predefined constant values for the following command strings:
   CMDSTR_NEWFOLDER	"NewFolder"
   CMDSTR_VIEWDETAIL	"ViewDetails"
   CMDSTR_VIEWLIST	"ViewList"
lpParameters   Optional parameters.
               This member is always NULL for menu items inserted by a shell extension.
lpDirectory    Optional working directory name.
               This member is always NULL for menu items inserted by a shell extension.
nShow          Flag to pass to the ShowWindow function if the command displays a window or starts an application.
dwHotKey       Optional hot key to assign any application activated by the command.
               If fMask does not specify CMIC_MASK_HOTKEY, this member is ignored.
hIcon          Icon to use for any application activated by the command.
               If fMask does not specify CMIC_MASK_ICON, this member is ignored.
}
var
   IDOffset : Integer;
begin
{$IFDEF Debug}
Loggit('InvokeCommand Begins');
{$ENDIF}
   Result := E_INVALIDARG;
   if HiWord(Integer(lpici.lpVerb)) <> 0 then begin
{$IFDEF Debug}
Loggit('InvokeCommand called by application');
{$ENDIF}
      //We are being called by an application:
      //Find the menu by its Verb (Name)
      IDOffset := FContextMenu.IDOfString(String(PChar(lpici.lpVerb)^), cmstVerb);
   end else begin
{$IFDEF Debug}
Loggit('InvokeCommand called by Shell');
{$ENDIF}
      //We are being called by the Shell
      IDOffset := LoWord(lpici.lpVerb);
   end; { else }
   if IDOffset < 0 then Exit; //Unrecognized Verb

   FhWnd := lpici.hWnd;
   //Execute the command specified
   try
      if Assigned(FContextMenu.OnClickByID[IDOffset]) then FContextMenu.OnClickByID[IDOffset](FContextMenu);
      Result := NOERROR;
   except
      Result := E_FAIL;
   end; { try }

   //Context menu will be re-Initialized before we reach InvokeCommand again,
   //so we might as well trash the FFiles and FContextMenu.
   FFiles.Free;
   FContextMenu.Free;
{$IFDEF Debug}
Loggit('InvokeCommand Ends');
{$ENDIF}
end; { IContextMenu.InvokeCommand }

function TCOMObjContextMenu.HandleMenuMsg(uMsg : UINT; wParam : WPARAM; lParam : LPARAM) : HResult;
{
   Invoked when the context menu needs to handle the WM_INITMENUPOPUP, WM_DRAWITEM
   and WM_MEASUREITEM messages (the menu involves owner-drawn menu items).
}
begin
{$IFDEF Debug}
Loggit('HandleMenuMsg');
{$ENDIF}
   Result := NOERROR;
   case uMsg of
   WM_INITMENUPOPUP : begin
      try
         DoInitMenuPopup(wParam, LoWord(lParam), HiWord(lParam) = 0);
      except
         Result := E_FAIL;
      end;
   end; { WM_INITMENUPOPUP }

   WM_MEASUREITEM : begin
      try
         DoMeasureItem(PMeasureItemStruct(lParam)^);
      except
         Result := E_FAIL;
      end; { try }
   end; { WM_MEASUREITEM }

   WM_DRAWITEM : begin
      try
         DoDrawItem(PDrawItemStruct(lParam)^);
      except
         Result := E_FAIL;
      end; { try }
   end; { WM_DRAWITEM }

// else
   end; { case }

(*
   try
      Result := DoMenuMsg(uMsg, wParam, lParam);
   except
      Result := E_FAIL;
   end; { try }
*)
end; { IContextMenu2.HandleMenuMsg }

function TCOMObjContextMenu.HandleMenuMsg2(uMsg : UINT; wParam : WPARAM; lParam : LPARAM; var MsgResult : LongInt) : HResult;
{
   Invoked when the context menu needs to handle the WM_MENUCHAR message
   (typically when an accelerator is used to select a menu item that displays a bitmap).
   WM_MENUCHAR is sent when the user presses a key that does not correspond to any
   mnemonic or accelerator key; an application should process this message when an
   accelerator is used to select a menu item that displays a bitmap.
   Before calling the handler, we attempt to match the caption's accelerator key;
   you should rarely have to override DoMenuMsg in that case.
}
var
   NMatches,
   MenuIndex,
   Index : Integer;
begin
{$IFDEF Debug}
Loggit('HandleMenuMsg2');
{$ENDIF}
// case uMsg of
// WM_MENUCHAR : begin
      //HiWord(wParam) will always be MF_POPUP
      { Any caption matches? }
      NMatches := 0;
      for Index := 0 to (FContextMenu.Count - 1) do begin
         if FContextMenu.HasMenuBitmap[Index] and
            IsAccel(LoWord(wParam), FContextMenu.Captions[Index]) then begin
            Inc(NMatches);
            { Found a match: What's the entry's position? }
            for MenuIndex := 0 to (GetMenuItemCount(lParam) - 1) do begin
               if (FContextMenu.IDOffset + FContextMenu.ID[Index]) = GetMenuItemID(lParam, MenuIndex) then begin
                  { Found the entry's menu position }
                  if NMatches = 1 then begin
                     { First match: Activate the menu entry }
                     MsgResult := (MNC_EXECUTE shl 16) or MenuIndex;
                  end else begin
                     { Multiple matches }
                     if (MenuIndex > GetHighlightedMenuItem(lParam)) and
                        (LoWord(MsgResult) <= GetHighlightedMenuItem(lParam)) then begin
                        { Move focus to the new menu entry }
                        MsgResult := (MNC_SELECT shl 16) or MenuIndex;
                     end else begin
                        { Move focus to the old menu entry }
                        MsgResult := (MNC_SELECT shl 16) or LoWord(MsgResult);
                     end; { else }
                  end; { else }
                  break;
               end; { then }
            end; { for }
            //We must check all of our entries for multiple caption hits (otherwise we'd Break here)
            {
            If none of the shell's own menu entries recognize the accelerator,
            the shell applies this method to each loaded extension until it gets
            a hit --note that the sequence of extensions is unknown to us.
            We cannot poll the other extensions' MF_BITMAP menus ourselves (and
            we have no access to their captions, if any), so we cannot know
            whether there'll be multiple hits or not.
            }
         end; { then }
      end; { for }
      MenuIndex := MsgResult and $0000FFFF;
      Index := MsgResult shr 16; { Action }
      try
         DoMenuChar(Char(LoWord(wParam)), lParam, MenuIndex, Index);
         MsgResult := (Index shl 16) or Index;
         Result := NOERROR;
      except
         Result := E_FAIL;
      end; { try }
// end; { WM_MENUCHAR }
end; { IContextMenu3.HandleMenuMsg2 }

function TCOMObjContextMenu.DoQuery : HResult;
begin
   Result := S_OK;
end; { TCOMObjContextMenu.DoQuery }

function TCOMObjContextMenu.DoMenuMsg(uMsg : UINT; wParam : WPARAM; lParam : LPARAM) : HResult;
begin
   Result := S_OK;
end; { TCOMObjContextMenu.DoMenuMsg }

procedure TCOMObjContextMenu.DoInitMenuPopup(Menu : HMENU; Index : Integer; SystemMenu : Boolean);
begin
   //No action
{$IFDEF Debug}
if SystemMenu then
Loggit('DoInitMenuPopup: Menu = ' + IntToStr(Menu) + ' Index = ' + IntToStr(Index) + ' SystemMenu = True')
else
Loggit('DoInitMenuPopup: Menu = ' + IntToStr(Menu) + ' Index = ' + IntToStr(Index) + ' SystemMenu = False');
{$ENDIF}
end; { TCOMObjContextMenu.DoInitMenuPopup }

procedure TCOMObjContextMenu.DoMeasureItem(var MeasureItem : TMeasureItemStruct);
(*
  TMeasureItemStruct = packed record
    CtlType: UINT;
    CtlID: UINT;
    itemID: UINT;
    itemWidth: UINT;
    itemHeight: UINT;
    itemData: DWORD;
  end;
*)
begin
   //No action
{$IFDEF Debug}
Loggit('DoMeasureItem');
{$ENDIF}
end; { TCOMObjContextMenu.DoMeasureItem }

procedure TCOMObjContextMenu.DoDrawItem(var DrawItem : TDrawItemStruct);
(*
  TDrawItemStruct = packed record
    CtlType: UINT;
    CtlID: UINT;
    itemID: UINT;
    itemAction: UINT;
    itemState: UINT;
    hwndItem: HWND;
    hDC: HDC;
    rcItem: TRect;
    itemData: DWORD;
  end;
*)
begin
   //No action
{$IFDEF Debug}
Loggit('DoDrawItem');
{$ENDIF}
end; { TCOMObjContextMenu.DoDrawItem }

procedure TCOMObjContextMenu.DoMenuChar(Key : Char; Menu : HMENU; var Index, Action : Integer);
begin
   //No action
end; { TCOMObjContextMenu.DoMenuChar }

end. { unit COMObjContextMenu }
