unit COMObjCopyHook;
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   16 September 1998
   Updated 17 February 1999
    1998-1999 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use the 
   component or any part of its code, you must include an 
   acknowledgement in the finished code or executable (in the About 
   box and/or the Help file).

   TCOMObjCopyHook:
   Generic TCOMObject descendant that implements the ICopyHook interface.
   It is designed so the user need only override the DoCopyHook function.
*)
interface

uses
   Windows,
   Classes, { TStringList }
   COMObj, { TCOMObject, etc. }
   ShlObj; { ICopyHook, etc. }

const
   { Additional file attribute flags missed by SysUtils }
   faNormal     = $00000080;
   faTemporary  = $00000100;
   faCompressed = $00000800;
   faOffLine    = $00001000;

type
   TCopyHookOperation = (chFileCopy, chFileDelete, chFileMove, chFileRename,
                         chPrinterDelete, chPrinterPortChange, chPrinterRename,
                         chPrinterRenameAndPortChange);
{
   chFileCopy	         The file specified by FSource will be copied to
                        the location specified by FDest.
   chFileDelete	      The file specified by FSource will be deleted.
   chFileMove	         The file specified by FSource will be moved to
                        the location specified by FDest.
   chFileRename	      The file specified by FSource will be renamed to FDest.
   chPrinterDelete      The printer specified by FSource will be deleted.
   chPrinterPortChange  The printer port is about to change.
                        FSourcePrinters contains the affected printer names,
                        FSourcePorts contains the current printer ports, and
                        FDestPorts contains the proposed new printer ports.
   chPrinterRename	   The printer specified by FSource will be renamed to FDest.
   chPrinterRenameAndPortChange	   Combination of chPrinterRename and chPrinterPortChange;
                        in this case FDestPrinters contains the proposed new printer names.
}

   TCOMObjCopyHook = class(TCOMObject, ICopyHook)
   protected
   { Available to descendants }
      FhWnd          : HWND;
      {
      Handle of the window that the copy hook handler should use
      as the parent window for any user interface elements the
      handler may need to display.
      If FOF_SILENT is specified in FFlags, DoCopyHook should ignore this parameter.
      }
      FOperation     : TCopyHookOperation;
      {
      Specifies the type of operation about to occur.
      }
      FFlags         : UINT;
      {
      FOF_ALLOWUNDO	         Undo information will be preserved if possible.
      FOF_CONFIRMMOUSE	      Not implemented.
      FOF_FILESONLY	         Not implemented.  The shell calls a
         copy hook handler only for folder objects, not files.
      FOF_MULTIDESTFILES	   The SHFileOperation function specified
         multiple destination files (one for each source file) rather
         than one directory where all the source files are to be deposited.
         A copy hook handler typically ignores this value since it is
         prompted on a file-by-file basis.
      FOF_NOCONFIRMATION	   Responds with "yes to all" to any dialog box
         that is displayed.  (Which means what, exactly?)
      FOF_NOCONFIRMMKDIR	   The shell won't ask for permission to create
         any needed directories.
      FOF_RENAMEONCOLLISION	The file being operated on will be renamed
         in a copy, move, or rename operation if a file of the target name
         already exists (e.g. "Copy #1 of ...").
      FOF_SILENT	            No progress dialog box will be displayed.
      FOF_SIMPLEPROGRESS	   The progress dialog box does not show the
         names of the files.
      }
      FSource,
      FDest          : String;
      {
      The source and destination files.  See TCopyHookOperation for details.
      }
      FSourcePrinters,
      FDestPrinters,
      {
      The source and destination printer names.  See TCopyHookOperation for details.
      }
      FSourcePorts,
      FDestPorts     : TStringList;
      {
      The source and destination printer ports.  See TCopyHookOperation for details.
      }
      FSourceAttribs,
      FDestAttribs   : Integer;
      {
      The source and destination file attributes.  See TCopyHookOperation for details.
      }

      function DoCopyHook : UINT; virtual;
      {
      Designed to be overridden, DoCopyHook returns one of:
      IDYES	   Allows the operation.
      IDNO	   Prevents the operation on this file, but continues with
         any other operations (for example, a batch copy operation).
      IDCANCEL Prevents the current operation and cancels any pending
         operations.
      The default implementation simply returns IDYES.
      }
   public
   { Available at run-time }
      function CopyCallback(Wnd : HWND; wFunc, wFlags : UINT; pszSrcFile : PAnsiChar; dwSrcAttribs : DWORD; pszDestFile : PAnsiChar; dwDestAttribs : DWORD) : UINT; stdcall;
   end;

implementation

uses
   ShellAPI; { FO_Copy, etc. }

{ ########## TCOMObjCopyHook ########## }

function TCOMObjCopyHook.CopyCallback(Wnd : HWND; wFunc, wFlags : UINT; pszSrcFile : PAnsiChar; dwSrcAttribs : DWORD; pszDestFile : PAnsiChar; dwDestAttribs : DWORD) : UINT;
{
Wnd
   Handle of the window that the copy hook handler should use
   as the parent window for any user interface elements the
   handler may need to display.
   If FOF_SILENT is specified, the method should ignore this parameter.

wFunc
   Operation about to be performed.
   This parameter can be one of the following values:
   FO_COPY	      The file specified by pszSrcFile will be copied to
      the location specified by pszDestFile.
   FO_DELETE	   The file specified by pszSrcFile will be deleted.
   FO_MOVE	      The file specified by pszSrcFile will be moved to
      the location specified by pszDestFile.
   FO_RENAME	   The file specified by pszSrcFile will be renamed.
   PO_DELETE	   The printer specified by pszSrcFile will be deleted.
   PO_PORTCHANGE  The printer port is about to change.
      The pszSrcFile and pszDestFile parameters contain double
      null-terminated lists of strings.  Each list contains
      printer names followed by port names.  The port name in
      pszSrcFile is the current printer port, and the port name in
      pszDestFile is the new printer port.
   PO_RENAME	   The printer specified by pszSrcFile will be renamed.
   PO_REN_PORT	   Combination of PO_RENAME and PO_PORTCHANGE.

wFlags
   Flags that will control the operation.
   This parameter can be a combination of the following values:
   FOF_ALLOWUNDO	         Preserves undo information (if possible).
   FOF_CONFIRMMOUSE	      Not implemented.
   FOF_FILESONLY	         Not implemented.  The shell calls a
      copy hook handler only for folder objects, not files.
   FOF_MULTIDESTFILES	   The SHFileOperation function specified
      multiple destination files (one for each source file) rather
      than one directory where all the source files are to be deposited.
      A copy hook handler typically ignores this value.
   FOF_NOCONFIRMATION	   Responds with "yes to all" to any dialog box
      that is displayed.
   FOF_NOCONFIRMMKDIR	   Does not confirm the creation of any needed
      directories if the operation requires a new directory to be created.
   FOF_RENAMEONCOLLISION	Gives the file being operated on a new name
      (such as, "Copy #1 of ...") in a copy, move, or rename operation
      if a file of the target name already exists.
   FOF_SILENT	            Displays no progress dialog box.
   FOF_SIMPLEPROGRESS	   Displays a progress dialog box, but the dialog
      box does not show the names of the files.

pszSrcFile
   Pointer to a string that contains the name of the source file.

dwSrcAttribs
   Attributes of the source file.
   This parameter can be a combination of any of the file attribute flags
   (FILE_ATTRIBUTE_*) defined in the Windows header files.

pszDestFile
   Pointer to a string that contains the name of the destination file.

dwDestAttribs
   Attributes of the destination file.
   This parameter can be a combination of any of the file attribute flags
   (FILE_ATTRIBUTE_*) defined in the Windows header files.

Return Value
   Returns an integer value that indicates whether the shell should
   perform the operation.  It can be one of the following values:
   IDYES	   Allows the operation.
   IDNO	   Prevents the operation on this file, but continues with
      any other operations (for example, a batch copy operation).
   IDCANCEL Prevents the current operation and cancels any pending
      operations.
}
var
   P : PChar;
begin
   FhWnd := Wnd;
   case wFunc of
   FO_COPY        : FOperation := chFileCopy;
   FO_DELETE      : FOperation := chFileDelete;
   FO_MOVE        : FOperation := chFileMove;
   FO_RENAME      : FOperation := chFileRename;
   PO_DELETE      : FOperation := chPrinterDelete;
   PO_PORTCHANGE  : FOperation := chPrinterPortChange;
   PO_RENAME      : FOperation := chPrinterRename;
   PO_REN_PORT    : FOperation := chPrinterRenameAndPortChange;
   end; { case }
   FFlags := wFlags;
   FSourcePrinters := TStringList.Create;
   FDestPrinters   := TStringList.Create;
   FSourcePorts    := TStringList.Create;
   FDestPorts      := TStringList.Create;
   if wFunc in [PO_PORTCHANGE, PO_REN_PORT] then begin
      FSource := '';
      FDest   := '';
      P := pszSrcFile;
      repeat
         FSourcePrinters.Append(String(P));
         P := PChar(Integer(P) + Length(String(P)) + 1);
         FSourcePorts.Append(String(P));
         P := PChar(Integer(P) + Length(String(P)) + 1);
      until Length(String(P)) = 0;
      { Presumably pszDestFile is in the same sequence as pszSrcFile }
      P := pszDestFile;
      repeat
         FDestPrinters.Append(String(P));
         P := PChar(Integer(P) + Length(String(P)) + 1);
         FDestPorts.Append(String(P));
         P := PChar(Integer(P) + Length(String(P)) + 1);
      until Length(String(P)) = 0;
   end else begin
      FSource := String(pszSrcFile);
      FDest   := String(pszDestFile);
   end; { else }
   FSourceAttribs := dwSrcAttribs;
   FDestAttribs   := dwDestAttribs;

   try
      Result := DoCopyHook;
   except
      Result := IDCANCEL;
   end; { try }

   FSourcePrinters.Free;
   FDestPrinters.Free;
   FSourcePorts.Free;
   FDestPorts.Free;
end; { ICopyHook.CopyCallback }

function TCOMObjCopyHook.DoCopyHook : UINT;
begin
   Result := IDYES;
end; { TCOMObjCopyHook.DoCopyHook }

end. { unit COMObjCopyHook }
