unit COMObjExtractIcon;
{..$DEFINE Debug}
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   30 January 1999
   Updated 17 February 1999
    1999 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use the 
   component or any part of its code, you must include an 
   acknowledgement in the finished code or executable (in the About 
   box and/or the Help file).

   TCOMObjExtractIcon:
   A generic IExtractIcon implementation.
   It is designed so the user need only override the
   DoGetIconLocation and DoExtractIcon methods.
*)
interface

uses
   Windows,
   COMObjPersistFile, { TCOMObjPersistFile }
   ShlObj; { IExtractIcon }

type
   TCOMObjExtractIcon = class(TCOMObjPersistFile, IExtractIcon)
   protected
   { Available to descendants }
      function DoGetIconLocation(IconFileLength, IconFlags : UINT; var IconFile : String; var Index, Flags : UINT) : HRESULT; virtual;
      {
      IconFileLength    Maximum length of the returned IconFile (almost always MAX_PATH).
      IconFlags         This input can be zero or either of:
         GIL_FORSHELL	The icon is to be displayed in a shell folder.
         GIL_OPENICON	The icon is for a folder that is open.
      IconFile          The method should set this to the icon location fully qualified filename.
      Index             The method should set this to the icon index into the IconFile (zero-based).
      Flags             The method should set this to zero or more of the following values:
         GIL_DONTCACHE	   The icon should not be cached.
         GIL_NOTFILENAME	The location is not a filename.
                           This will cause a call to the IExtractIcon::Extract method.
         GIL_PERCLASS	   All file objects of this class have the same icon.
         GIL_PERINSTANCE	Each file object of this class has its own icon.
         GIL_SIMULATEDOC	The shell should create a document icon using the specified icon.
      Result:
      NOERROR, if a valid location has been set;
      S_FALSE, if the shell should use a default icon.

      The default implementation simply returns S_FALSE.
      }
      function DoExtractIcon(IconFile : String; Index : UINT; SizeBig, SizeSmall : Word; var hIconBig, hIconSmall : hIcon) : HRESULT; virtual;
      {
      IconFile    The icon file location or some other meaningful instructions.
                  This value matches the IconFile returned by DoGetIconLocation.
      Index       Icon index.
                  This value matches the Index returned by DoGetIconLocation.
      SizeBig     The size, in pixels, of the large icon required (typically 32 or 48).
      SizeSmall   The size, in pixels, of the small icon required (typically 16).
      hIconBig    Output handle of the large icon.
      hIconSmall  Output handle of the small icon.
      Result:
      NOERROR, if DoExtractIcon successfully extracted or generated the icon.
      S_FALSE, if the caller should extract the icon from the file specified
         by the IconFile and IconIndex values instead.

      The default implementation simply returns S_FALSE.
      }
   public
   { Available at run-time }
      //IExtractIcon
      function GetIconLocation(uFlags : UINT; szIconFile : PAnsiChar; cchMax : UINT; out piIndex : Integer; out pwFlags : UINT) : HResult; stdcall;
      function Extract(pszFile : PAnsiChar; nIconIndex : UINT; out phiconLarge, phiconSmall : HICON; nIconSize : UINT) : HResult; stdcall;
   end;
(*
uses Graphics; { TBitmap }
   function CreateIconFromPicts(var PictBmp, MaskBmp : Graphics.TBitmap) : hIcon;
*)
implementation

uses
{$IFDEF Debug}
DLLDebug, { Loggit }
{$ENDIF}
   SysUtils; { StrLCopy, PByteArray }
(*
function CreateIconFromPicts(var PictBmp, MaskBmp : Graphics.TBitmap) : hIcon;
{
   Creates an icon from two bitmaps.
PictBmp  The icon bitmap
MaskBmp  The mask bitmap; this MUST be Monochrome.
         If Nil, a mask is created from the PictBmp
         by using the top-left pixel colour as background.
}
var
   vTB     : Windows.TBitmap;
   PictSiz,
   MaskSiz : Integer;
   NewPict,
   NewMask : PByteArray;
begin
   //Return a null hIcon in case of failure
   Result := 0;

   if not Assigned(PictBmp) then Exit;
   //Create MaskBmp if required
   if not Assigned(MaskBmp) then begin
      MaskBmp := TBitmap.Create;
      with MaskBmp do begin
         Assign(PictBmp);
         Dormant;   //Free up GDI resources
         FreeImage; //Free up Memory.
         //Size is now set
         Monochrome := True;
         Transparent := True;
         TransParentColor := PictBmp.Canvas.Pixels[0, 0];
         Canvas.Draw(0, 0, PictBmp);
      end; { with }
   end; { then }

   if not MaskBmp.Monochrome then exit;

   with PictBmp do begin
      GetObject(Handle, SizeOf(vTB), @vTB);
      PictSiz := (Height * Width * vTB.bmBitsPixel * vTB.bmPlanes) div 8;
      MaskSiz := (Height * Width) div 8; //Mask is always two-color
   end; { with }

   GetMem(NewPict, PictSiz);
   GetMem(NewMask, MaskSiz);
   try
      GetBitmapBits(PictBmp.Handle, PictSiz, NewPict);
      GetBitmapBits(MaskBmp.Handle, MaskSiz, NewMask);
      Result := CreateIcon(hInstance, vTB.bmWidth, vTB.bmHeight, vTB.bmPlanes, vTB.bmBitsPixel, NewMask, NewPict);
   finally
      FreeMem(NewPict, PictSiz);
      FreeMem(NewMask, MaskSiz);
   end; { try }
end; { CreateIconFromPicts }
*)
{ ########## TCOMObjExtractIcon ########## }

function TCOMObjExtractIcon.GetIconLocation(uFlags : UINT; szIconFile : PAnsiChar; cchMax : UINT; out piIndex : Integer; out pwFlags : UINT) : HResult; stdcall;
{
   This function returns an icon location.

   Parameters:
   uFlags     This input can be zero or either of:
      GIL_FORSHELL	The icon is to be displayed in a shell folder.
      GIL_OPENICON	The icon is for a folder that is open.
   szIconFile Pointer to the string buffer that will receive the icon location filename.
   cchMax     Size of szIconFile (almost always MAX_PATH).
   piIndex    Pointer to an integer that will receive the icon index.
   pwFlags    Pointer to an unsigned integer that receives zero or more
      of the following values:
      GIL_DONTCACHE	   The icon should not be cached.
      GIL_NOTFILENAME	The location is not a filename.
                        This will cause a call to the IExtractIcon::Extract method.
      GIL_PERCLASS	   All file objects of this class have the same icon.
      GIL_PERINSTANCE	Each file object of this class has its own icon.
      GIL_SIMULATEDOC	The shell should create a document icon using the specified icon.
   Returns:
   NOERROR, if it returns a valid location;
   S_FALSE, if the shell should use a default icon.

   When GIL_NOTFILENAME is set, szIconFile will be passed verbatim to Extract.
}
var
   IconFile : String;
begin
   try
      Result := DoGetIconLocation(cchMax - 1, uFlags, IconFile, piIndex, pwFlags);
      if Result = NOERROR then StrLCopy(szIconFile, PChar(IconFile), cchMax);
   except
      Result := S_FALSE;
   end; { try }
end; { IExtractIcon.GetIconLocation }

function TCOMObjExtractIcon.Extract(pszFile : PAnsiChar; nIconIndex : UINT; out phiconLarge, phiconSmall : HICON; nIconSize : UINT) : HResult; stdcall;
{
   This function extracts an icon image from a specified file.

   Parameters:
   pszFile     Pointer to a null-terminated string specifying the
      icon location or some other meaningful instructions.
      The caller should set pszFile to the szIconFile returned by GetIconLocation.
   nIconIndex  Icon index.
      The caller should set nIconIndex to the piIndex returned by GetIconLocation.
   phiconLarge Pointer to an HIcon that receives the handle of the large icon
   phiconSmall Pointer to an HIcon that receives the handle of the small icon
   nIconSize   Specifies the size, in pixels, of the large icon required
      LOWORD is the requested large icon size
		HIWORD is the requested small icon size (this wasn't in the Win32 doc!)
   Returns:
   NOERROR, if Extract successfully extracted or generated the icon.
   S_FALSE, if the caller should extract the icon from the file specified
      by the pszFile/szIconFile and nIconIndex/piIndex values.
}
begin
   try
      Result := DoExtractIcon(String(pszFile), nIconIndex, LoWord(nIconSize), HiWord(nIconSize), phiconLarge, phiconSmall);
   except
      Result := S_FALSE;
   end; { try }
end; { IExtractIcon.Extract }

function TCOMObjExtractIcon.DoGetIconLocation(IconFileLength, IconFlags : UINT; var IconFile : String; var Index, Flags : UINT) : HRESULT;
begin
   Result := S_FALSE;
end; { TCOMObjExtractIcon.DoGetIconLocation }

function TCOMObjExtractIcon.DoExtractIcon(IconFile : String; Index : UINT; SizeBig, SizeSmall : Word; var hIconBig, hIconSmall : hIcon) : HRESULT;
begin
   Result := S_FALSE;
end; { TCOMObjExtractIcon.DoExtractIcon }

end. { unit COMObjExtractIcon }
