unit COMObjPersistFile;
{..$DEFINE Debug}
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   30 January 1999
   Updated 17 February 1999
    1999 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use the 
   component or any part of its code, you must include an 
   acknowledgement in the finished code or executable (in the About 
   box and/or the Help file).

   TCOMObjPersistFile:
   Generic TCOMObject descendant that implements the IPersistFile interface.
   It is designed so the user need only override the Do... methods.
*)
interface

uses
   Windows,
   COMObj, { TCOMObject, etc. }
   ActiveX; { IPersist, IPersistFile, etc. }

type
   TCOMObjPersistFile = class(TCOMObject, IPersistFile)
   protected
      FFileName   : String;
      {
      Filename being acted upon.
      }
      FMode       : LongInt;
      {
      Specifies some combination of the values from the STGM enumeration
      to indicate the access mode to use when opening the file.
      DoLoad can treat this value as a suggestion,
      adding more restrictive permissions if necessary.
      If dwMode is zero, the implementation should open the file
      using whatever default permissions are used when a user opens the file.
      }
      FDirty,
      {
      The implementation's Dirty flag.
      Should be True whenever the memory object is out of sync with the file object.
      }
      FNoScribble : Boolean;
      {
      The implementation's NoScribble flag.
      It is set and reset by the Save and SaveCompleted methods;
      descendants will rarely need to access it.
      Whilst NoScribble is True, the object cannot write to its file.
      }
      function DoLoad : HResult; virtual;
      {
      Should return one of:
      S_OK        The object was successfully loaded.
      E_NOMEMORY  The object could not be loaded due to a lack of memory.
      E_FAIL      The object could not be loaded for some other reason.
      The default function simply returns S_OK.
      }
      function DoIsDirty : HResult; virtual;
      {
      Should return one of:
      S_OK     The object has changed since it was last saved.
      S_FALSE  The object has not changed since the last save.
      The default function simply returns S_FALSE.
      }
      function DoSave(FileName : String; Remember : BOOL) : HResult; virtual;
      {
      FileName
      The absolute path of the file to which the object should be saved.
      If FileName is empty, the object should save its data
      to the current file (FFileName), if there is one.
      Remember
      If FileName is empty, the implementation should ignore the Remember flag.
      If TRUE, FileName will become the current file
      and the object should clear its dirty flag after the save.
      If FALSE, this save operation is a "Save A Copy As ...".
      In this case, the current file is unchanged and the object
      should not clear its dirty flag.
      Should return one of:
      S_OK     The object was successfully saved.
      E_FAIL   The file was not saved.

      If FNoScribble was True, DoSave isn't called and Save returns E_FAIL.
      Otherwise, if DoSave returns S_OK, FNoScribble becomes True;
      in addition, if FileName was non-empty and Remember was True,
      FFileName is replaced by FileName and FDirty is set to False.

      The default function simply returns S_OK.
      }
      procedure DoSaveCompleted(FileName : String); virtual;
      {
      After each call to Save, the caller must eventually call SaveCompleted
      to clear the FNoScribble flag.
      The FileName passed in ought to match the FFileName
      (the documentation doesn't say what should happen if it does not).
      FNoScribble is cleared before DoSaveCompleted is called.
      The default procedure does nothing.
      }
      function DoGetCurFile : HResult; virtual;
      {
      Invoked once the memory has been allocated to return FFileName,
      this method allows the user a chance to modify FFileName before
      it is commited to the memory block and/or to deny GetCurFile's success.
      Should return one of:
      S_OK     The FFileName will be returned by GetCurFile
      S_FALSE  The FFileName will be returned by GetCurFile then reset to an empty string
               (DoGetCurFile should set FFileName to a default save prompt)
      E_FAIL   GetCurFile passes the result along.
      The default function returns S_OK if FFileName is non-empty,
      S_FALSE if it is.
      }
   public
      //IPersist (ancestor of IPersistFile)
      function GetClassID(out ClassID : TCLSID): HResult; stdcall;
      //IPersistFile
      function IsDirty : HResult; stdcall;
      function Load(pszFileName : POLEStr; dwMode : Longint) : HResult; stdcall;
      function Save(pszFileName : POLEStr; fRemember : BOOL) : HResult; stdcall;
      function SaveCompleted(pszFileName : POLEStr) : HResult; stdcall;
      function GetCurFile(out pszFileName : POLEStr) : HResult; stdcall;
   end;

implementation

{$IFDEF Debug}
uses
   DLLDebug; { Loggit }
{$ENDIF}

{ ########## TCOMObjPersistFile ########## }

function TCOMObjPersistFile.GetClassID(out ClassID : TCLSID): HResult;
{
   Returns the object's CLSID; we get it from the COMObjectFactory.
}
begin
   ClassID := Factory.ClassID;
   Result := S_OK;
end; { IPersist.GetClassID }

function TCOMObjPersistFile.Load(pszFileName : POLEStr; dwMode : Longint) : HResult;
{
   This is the entry point to the TCOMObjPersistFile.
   This implementation simply stores the filename (FFileName) and access mode (FMode)
   and then invokes DoLoad.
pszFileName
   Points to a zero-terminated string containing the absolute path of the file to open.
   (i.e. the file that is the current drop target)
dwMode
   Specifies some combination of the values from the STGM enumeration
   to indicate the access mode to use when opening the file.
   The IPersistFile::Load method can treat this value as a suggestion,
   adding more restrictive permissions if necessary.
   If dwMode is zero, the implementation should open the file
   using whatever default permissions are used when a user opens the file.
}
begin
   FFileName := WideCharToString(pszFilename);
   FMode := dwMode;
   FDirty := False;
   Result := DoLoad;
end; { IPersistFile.Load }

function TCOMObjPersistFile.IsDirty : HResult;
{
   Whether the object has changed since it was last saved to file.
}
begin
   Result := DoIsDirty;
end; { IPersistFile.IsDirty }

function TCOMObjPersistFile.Save(pszFileName : POLEStr; fRemember : BOOL) : HResult;
{
pszFileName
   Points to a zero-terminated string containing the absolute path
   of the file to which the object should be saved.
   If pszFileName is NULL, the object should save its data
   to the current file, if there is one.

fRemember
   Indicates whether the pszFileName parameter is to be used
   as the current working file.  If TRUE, pszFileName becomes
   the current file and the object should clear its dirty flag
   after the save.  If FALSE, this save operation is a
   "Save A Copy As ..." operation.  In this case, the
   current file is unchanged and the object should not clear
   its dirty flag.
   If pszFileName is NULL, the implementation should ignore the fRemember flag.
}
var
   FileName : String;
begin
   if FNoScribble then begin
      Result := E_FAIL;
   end else begin
      FileName := WideCharToString(pszFilename);
      Result := DoSave(Filename, fRemember);
      if Result = S_OK then begin
         FNoScribble := True;
         if (FileName <> '') and fRemember then begin
            FFileName := FileName;
            FDirty := False;
         end; { then }
      end; { then }
   end; { else }
end; { IPersistFile.Save }

function TCOMObjPersistFile.SaveCompleted(pszFileName : POLEStr) : HResult;
{
   Clients that call the Save method must follow it with a SaveCompleted call;
   the object is in NoScribble mode in between.
pszFileName
   Points to the absolute path of the file where the object was previously saved.
   (What happens if it doesn't match?  Dunno)
}
begin
   FNoScribble := False;
   DoSaveCompleted(WideCharToString(pszFileName));
   Result := S_OK;
end; { IPersistFile.SaveCompleted }

function TCOMObjPersistFile.GetCurFile(out pszFileName : POLEStr) : HResult;
{
   Return the current file storing the object.
pszFileName
   A pointer to be set to an IMalloc-obtained block, and that this
   method fills with the zero-terminated current file name.
   The caller is responsible for freeing the block using the same IMalloc.
}
var
   ppMalloc : IMalloc;
begin
   Result := E_FAIL;
   if S_OK <> CoGetMalloc(MEMCTX_TASK, ppMalloc) then Exit;
   Result := E_OUTOFMEMORY;
   pszFileName := ppMalloc.Alloc(2*MAX_PATH);
   if not Assigned(pszFileName) then Exit;
   Result := DoGetCurFile;
   if (Result = S_OK) or (Result = S_FALSE) then begin
      StringToWideChar(FFileName, pszFileName, MAX_PATH);
      if Result = S_FALSE then FFileName := '';
   end else begin
      ppMalloc.Free(pszFileName);
      pszFileName := Nil;
   end; { else }
end; { IPersistFile.GetCurFile }

function TCOMObjPersistFile.DoLoad : HResult;
begin
   Result := S_OK;
end; { TCOMObjPersistFile.DoLoad }

function TCOMObjPersistFile.DoIsDirty : HResult;
begin
   Result := S_FALSE;
end; { TCOMObjPersistFile.DoIsDirty }

function TCOMObjPersistFile.DoSave(FileName : String; Remember : BOOL) : HResult;
begin
   Result := S_OK;
end; { TCOMObjPersistFile.DoSave }

procedure TCOMObjPersistFile.DoSaveCompleted(FileName : String);
begin
end; { TCOMObjPersistFile.DoSaveCompleted }

function TCOMObjPersistFile.DoGetCurFile : HResult;
begin
   if FFileName = '' then begin
      Result := S_FALSE;
   end else begin
      Result := S_OK;
   end; { else }
end; { TCOMObjPersistFile.DoGetCurFile }

end. { unit COMObjPersistFile }
