unit COMObjShellExtInit;
{..$DEFINE Debug}
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   3 February 1999
   Updated 17 February 1999
    1999 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use the 
   component or any part of its code, you must include an 
   acknowledgement in the finished code or executable (in the About 
   box and/or the Help file).

   TCOMObjShellExtInit:
   Generic TCOMObject descendant that implements the IShellExtInit interface.
   It is designed so the user need only override the DoInitialize method as required.
*)
interface

uses
   Classes, { TStringList }
   Windows,
   COMObj,  { TCOMObject, etc. }
   ShlObj,  { IShellExtInit, etc. }
   ActiveX; { IDataObject, etc. }

type
   TCOMObjShellExtInit = class(TCOMObject, IShellExtInit)
   protected
   { Available to descendants }
      FPIDL       : PItemIDList;
      {
      Pointer to an ITEMIDLIST (item identifier list) that uniquely identifies a folder.
      Specifies the target folder for drag-and-drop extensions,
      Specifies the source folder for context menu and property sheet extensions.
      }
      FDataObj    : IDataObject;
      {
      Pointer to an IDataObject interface object that can be used to retrieve the objects being acted upon.
      }
      FhKeyProgID : HKEY;
      {
      Registry key specifying the file class of the file object that has the focus
      or the file class of the destination folder (drag-and-drop extensions).
      }
      FFiles      : TStringList;
      {
      The list of files being acted upon, when IDataObject can be rendered so, empty otherwise.
      }
      FIDPath     : String;
      {
      The path equivalent to FPIDL, when it exists.
      }

      function DoInitialize : HResult; virtual;
      {
      Called by Initialize once the protected Fields have been set up.
      Its Result is passed through by Initialize.
      Return NOERROR (a.k.a. S_OK) or some error HResult.
      The default implementation simply returns S_OK.
      }
   public
   { Available at run-time }
      //IShellExtInit
      //NOTE: This method hides the inherited one because its declaration is different.
      function Initialize(pidlFolder : PItemIDList; lpdobj : IDataObject; hKeyProgID : HKEY) : HResult; stdcall;
   end;

implementation

uses
{$IFDEF Debug}
DLLDebug, { Loggit }
{$ENDIF}
   SysUtils, { StrCopy, StrCat }
   ShellAPI; { DragQueryFile, etc. }

{ ########## TCOMObjShellExtInit ########## }

function TCOMObjShellExtInit.Initialize(pidlFolder : PItemIDList; lpdobj : IDataObject; hKeyProgID : HKEY) : HResult;
{
   Initialize is called first by the shell once the descendant extension is instantiated.
   This implementation simply stores the input parameters,
   retrieves the list of file names to act upon from lpdObj (FFiles) and
   translates pidlFolder into a path (FIDPath).
   You'll rarely need to do ought else.

pidlFolder
   PIDL (pointer to an ITEMIDLIST (item identifier list) structure that uniquely identifies a folder.
   For non-default drag and drop menu extensions, this parameter must specify the target folder.
   For context menu and property sheet extensions, specifies the source folder.
lpdobj
   Pointer to an IDataObject interface object that can be used to retrieve the objects being acted upon.
hKeyProgID
   Registry key specifying the file class of the file object that has the focus
   or the file class of the destination folder (drag-and-drop extensions).
}
var
   medium : TStgMedium;
   fe     : TFormatEtc;
   Index  : Integer;
   szFile : array[0..MAX_PATH] of Char;
begin
{$IFDEF Debug}
Loggit('TCOMObjShellExtInit.Initialize');
{$ENDIF}
   FPIDL := pidlFolder;
   FDataObj := lpdobj;
   FhKeyProgID := hKeyProgID;
   FFiles.Free;
   FFiles := TStringList.Create;
   FIDPath := '';

   SHGetPathFromIDList(pidlFolder, szFile);
   FIDPath := String(szFile);

   //Fail the call if lpdobj is Nil.  This should rarely occur.
   Result := E_FAIL;
   if lpdobj = Nil then Exit;

{
   Render the data referenced by the IDataObject pointer to
   an HGLOBAL storage medium in CF_HDROP format.
}
   with fe do begin
      cfFormat := CF_HDROP;
      ptd := Nil;
      dwAspect := DVASPECT_CONTENT;
      lindex := -1;
      tymed := TYMED_HGLOBAL;
   end;
   Result := lpdobj.GetData(fe, medium);
   if not Failed(Result) then begin
      //Retrieve the file name(s) and store it (them) in FFiles.
      for Index := 1 to DragQueryFile(medium.hGlobal, $FFFFFFFF, Nil, 0) do begin
         DragQueryFile(medium.hGlobal, (Index - 1), szFile, SizeOf(szFile));
         FFiles.Add(String(szFile));
      end; { for }
      ReleaseStgMedium(medium);
   end; { then }

   try
      Result := DoInitialize;
   except
      Result := E_FAIL;
   end; { try }
end; { IShellExtInit.Initialize }

function TCOMObjShellExtInit.DoInitialize : HResult;
begin
   Result := S_OK;
end; { TCOMObjShellExtInit.DoInitialize }

end. { unit COMObjShellExtInit }
