unit DUTShellDragDropHandler;
{..$DEFINE Debug}
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   30 January 1999
   Updated 17 February 1999
    1999 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use the 
   component or any part of its code, you must include an 
   acknowledgement in the finished code or executable (in the About 
   box and/or the Help file).

   The shell DragDropHandler extension.
   It is very similar to the ContextMenuHandler extension.

   A DragDropHandler is registered under the target file type,
   usually "Directory", "Drive" and/or "Folder" and is invoked when
   files are dragged to the target.  Since *any* collection of
   files may be so dragged, the first task of the QueryContextMenu
   should be to figure out whether the dragged set contains any
   files pertinent to its functionality.

   In this case,
   HKEY_CLASSES_ROOT\Directory\shellex\DragDropHandlers\DUTShellDragDropHandler = CLSID_DelDragDrop

   Not to be confused with DropHandlers (see the example).
*)
interface

uses
   Windows, { MessageBox }
   COMObj, { TComObjectFactory }
   COMObjContextMenu, { TCOMObjContextMenu }
   ShlObj; { SHChangeNotify, etc. }

const
   CLSID_DelDragDrop : TGUID = (
      D1:$6A5EF9E5; D2:$DFCE; D3:$11CF; D4:($9A, $34, $44, $45, $53, $54, $00, $00));

type
   TDragDropHandler = class(TCOMObjContextMenu)
   protected
   { Available to descendants }
      function DoQuery : HResult; override;
   public
   { Available at run-time }
      procedure CountFilesAndAdd(Sender : TObject);
      procedure CountFilesAndSet(Sender : TObject);
   end;

implementation

uses
{$IFDEF Debug}
DLLDebug, { Loggit }
{$ENDIF}
   SysUtils, { UpperCase, TSearchRec, etc. }
   COMServ, { COMServer }
   DUTDelShellUtils;

{ ########## TDragDropHandler ########## }

function TDragDropHandler.DoQuery : HResult;
var
   AllDelFiles : Boolean;
   Index       : Integer;
begin
{$IFDEF Debug}
Loggit('TDragDropHandler.DoQuery: Begins');
{$ENDIF}
   Result := inherited DoQuery; //Simply returns S_OK

   //Are we dropping a set consisting exclusively of DEL files?
   AllDelFiles := True;
   for Index := 0 to (FFiles.Count - 1) do begin
      if UpperCase(ExtractFileExt(FFiles[Index])) <> '.DEL' then begin
         AllDelFiles := False;
         Break;
      end; { then }
   end; { for }

   if AllDelFiles then begin
      //Add two menu items to context menu
      FContextMenu.Add('CountFilesAndAdd',
                       'Count Files and &Add',
                       'Count files in dropped-on folder and add to the magic numbers in dropped files',
                       CountFilesAndAdd,
                       False {Checked}, False {Grayed}, False {HasSeparator}, True {Visible});
      FContextMenu.Add('CountFilesAndSet',
                       'Count Files and Se&t',
                       'Count files in dropped-on folder and set the magic numbers of the dropped files to it',
                       CountFilesAndSet,
                       False {Checked}, False {Grayed}, True {HasSeparator}, True {Visible});
   end; { else }
// Result := S_OK;
{$IFDEF Debug}
Loggit('TDUTContextMenu.DoQuery: Ends');
{$ENDIF}
end; { TDragDropHandler.DoQuery }

procedure TDragDropHandler.CountFilesAndAdd(Sender : TObject);
var
   Rslt,
   Index,
   Count  : Integer;
   TS     : TSearchRec;
begin
{$IFDEF Debug}
Loggit('TDragDropHandler.CountFilesAndAdd');
{$ENDIF}
   //In this example, we count the number of files in the target folder
   //and add that value to each dropped file's "magic number".
   Count := 0;
   //Papparently, if FindFirst fails then FindClose should not be called
   Rslt := FindFirst(FIDPath + '\*.*', faReadOnly or faHidden or faSysFile or faArchive, TS);
   if Rslt = 0 then begin
      repeat
         Inc(Count);
         Rslt := FindNext(TS);
      until Rslt <> 0;
      FindClose(TS);
   end; { then }
   for Index := 0 to (FFiles.Count - 1) do begin
      IncDelFileNumber(FFiles[Index], Count);
      SHChangeNotify(SHCNE_UPDATEITEM, SHCNF_PATH or SHCNF_FLUSHNOWAIT, PChar(FFiles[Index]), Nil);
   end; { for }
   MessageBeep(0);
end; { TDragDropHandler.CountFilesAndAdd }

procedure TDragDropHandler.CountFilesAndSet(Sender : TObject);
var
   Rslt,
   Index,
   Count  : Integer;
   TS     : TSearchRec;
begin
{$IFDEF Debug}
Loggit('TDragDropHandler.CountFilesAndSet');
{$ENDIF}
   //In this example, we count the number of files in the target folder
   //and set each dropped file's "magic number" to that value.
   Count := 0;
   Rslt := FindFirst(FIDPath + '\*.*', faReadOnly or faHidden or faSysFile or faArchive, TS);
   while Rslt = 0 do begin
      Inc(Count);
      Rslt := FindNext(TS);
   end; { while }
   FindClose(TS);
   for Index := 0 to (FFiles.Count - 1) do begin
      SetDelFileNumber(FFiles[Index], Count);
      SHChangeNotify(SHCNE_UPDATEITEM, SHCNF_PATH or SHCNF_FLUSHNOWAIT, PChar(FFiles[Index]), Nil);
   end; { for }
   MessageBeep(0);
end; { TDragDropHandler.CountFilesAndSet }

initialization
   TComObjectFactory.Create(COMServer, TDragDropHandler, CLSID_DelDragDrop, '', 'DUTShellDragDropHandler', ciMultiInstance);

end. { unit DUTShellDragDropHandler }
