unit DUTShellDropHandler;
{..$DEFINE Debug}
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   30 January 1999
   Updated 17 February 1999
    1999 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use the 
   component or any part of its code, you must include an 
   acknowledgement in the finished code or executable (in the About 
   box and/or the Help file).

   The shell DropHandler extension.

   DropHandlers are registered under the target file type and
   are invoked when something is dropped onto the target file type.
   There can only be one DropHandler or IconHandler registered
   under any given target.
   By default, a file is not a drop target.  By providing a
   DropHandler for the file types created by your application,
   you can make the files into drop targets.  A DropHandler must
   support both the IPersistFile and IDropTarget interfaces.

   Explorer seems to check for DropHandlers only when it starts, so
   when you register a new one you need to restart Explorer.
   However, if the DropHandler was registered at Explorer start,
   unregistering it turns it off (as expected) but re-registering
   it turns it back on.  Go figure.

   The shell initializes a DropHandler by calling the
   IPersistFile::Load member function.
   When the user drags an object over one of your application's
   files or drops an object onto one of its files, the system
   calls the appropriate member functions of the IDropTarget interface.

   DropHandlers are mostly used by special system COM objects (registered
   under a CLSID) and by the "lnkfile" file type (shortcuts).
   For an example, see the "WinZip" file type.

   In this case,
   HKEY_CLASSES_ROOT\.del = DelShellFile
   HKEY_CLASSES_ROOT\DelShellFile\shellex\DropHandler = CLSID_DelDrop
*)
interface

uses
   Windows, { MessageBox }
   COMObj, { TComObjectFactory }
   DELCOMObjPersistFile, { TDELCOMObjPersistFile }
   ShlObj, { SHChangeNotify }
   ActiveX; { IDropTarget, etc. }

const
   CLSID_DelDrop : TGUID = (
      D1:$6A5EF9E6; D2:$DFCE; D3:$11CF; D4:($9A, $34, $44, $45, $53, $54, $00, $00));

type
   TDropHandler = class(TDELCOMObjPersistFile, IDropTarget)
   public
      //IDropTarget
      function DragEnter(const dataObj : IDataObject; grfKeyState : Longint; pt : TPoint; var dwEffect : Longint) : HResult; stdcall;
      function DragOver(grfKeyState : Longint; pt : TPoint; var dwEffect : Longint) : HResult; stdcall;
      function DragLeave : HResult; stdcall;
      function Drop(const dataObj : IDataObject; grfKeyState : Longint; pt : TPoint; var dwEffect : Longint) : HResult; stdcall;
   end;

implementation

uses
{$IFDEF Debug}
DLLDebug, { Loggit }
{$ENDIF}
   SysUtils, { IntToStr }
   ShellAPI, { DragQueryFile }
   COMServ, { COMServer }
   DUTDelShellUtils;

const
   { Declare a menu ID offset for each potential context menu item }
   IDOFFSET_COUNTITEM = 0;

{ ########## TDropHandler ########## }

function TDropHandler.DragEnter(const dataObj : IDataObject; grfKeyState : Longint; pt : TPoint; var dwEffect : Longint) : HResult;
{
   The DoDragDrop function calls this method the first time
   the user drags the mouse into the registered window of a drop target.
   (in this case, the icon and filename as exposed in a shell folder)

DataObj
   Points to the data object being transferred in the drag-and-drop operation.
   The data is represented in the IDataObject interface.
   If the drop occurs, this data object will be incorporated into the target.

grfKeyState
   Identifies the current state of the keyboard modifier keys on the keyboard.
   Valid values can be a combination of any of the flags
   MK_CONTROL, MK_SHIFT, MK_ALT, MK_BUTTON, MK_LBUTTON, MK_MBUTTON, and MK_RBUTTON.

pt
   Points to the current cursor coordinates in the coordinate space
   of the drop target window.

dwEffect
   Specifies the current effect flag.
   Valid values are from the DROPEFFECT enumeration:
   DROPEFFECT_NONE, DROPEFFECT_COPY, DROPEFFECT_MOVE, DROPEFFECT_LINK, DROPEFFECT_SCROLL
   The input value of pdwEffect establishes the possible effects; the method must choose
   one from those on output.
}
begin
   //Allow either Link or None.
   dwEffect := DROPEFFECT_LINK and dwEffect;
   Result := S_OK;
end; { IDropTarget.DragEnter }

function TDropHandler.DragOver(grfKeyState : Longint; pt : TPoint; var dwEffect : Longint) : HResult;
{
   This method is called as the drag proceeds over the target.
   It merely updates the DragEnter response.
   Note that the Shell already provides some visual feedback.
grfKeyState
   Identifies the current state of the keyboard modifier keys on the keyboard.
   Valid values can be a combination of any of the flags
   MK_CONTROL, MK_SHIFT, MK_ALT, MK_BUTTON, MK_LBUTTON, MK_MBUTTON, and MK_RBUTTON.

pt
   Points to the current cursor coordinates in the
   coordinate space of the drop target window.

dwEffect
   Specifies the current effect flag.
}
begin
   //Allow either Link or None.
   dwEffect := DROPEFFECT_LINK and dwEffect;
   Result := S_OK;
end; { IDropTarget.DragOver }

function TDropHandler.DragLeave : HResult;
{
   This method is called in either of the following cases:
	When the user drags the cursor out of a given target window.
	When the user cancels the current drag-and-drop operation.
   The method must remove any target feedback that is currently displayed.
   It must also release any references held to the data transfer object.
}
begin
   Result := S_OK;
end; { IDropTarget.DragLeave }

function TDropHandler.Drop(const dataObj : IDataObject; grfKeyState : Longint; pt : TPoint; var dwEffect : Longint) : HResult;
{
   This method is called when the actual Drop occurs.

DataObj
   Points to the data object being transferred in the drag-and-drop operation.
   The data is represented in the IDataObject interface.

grfKeyState
   Identifies the current state of the keyboard modifier keys on the keyboard.
   Valid values can be a combination of any of the flags
   MK_CONTROL, MK_SHIFT, MK_ALT, MK_BUTTON, MK_LBUTTON, MK_MBUTTON, and MK_RBUTTON.

pt
   Points to the current cursor coordinates
   in the coordinate space of the drop target window.

dwEffect
   Specifies the current effect flag.
}
var
   medium     : TStgMedium;
   fe         : TFormatEtc;
   NumDropped : Integer;
begin
   //Fail the call if DataObj is Nil.  This should rarely occur.
   Result := E_FAIL;
   if not Assigned(DataObj) then Exit;

{
   Render the data referenced by the IDataObject pointer to
   an HGLOBAL storage medium in CF_HDROP format.
}
   with fe do begin
      cfFormat := CF_HDROP;
      ptd := Nil;
      dwAspect := DVASPECT_CONTENT;
      lindex := -1;
      tymed := TYMED_HGLOBAL;
   end;
   Result := DataObj.GetData(fe, medium);
   if Failed(Result) then Exit;

   //Retrieve the number of files dropped.
   NumDropped := DragQueryFile(medium.hGlobal, $FFFFFFFF, Nil, 0);
   //Everything else is irrelevant.
   ReleaseStgMedium(medium);
   if SetDelFileNumber(FFileName, NumDropped) < 0 then begin
      Result := E_UNEXPECTED;
   end else begin
      SHChangeNotify(SHCNE_UPDATEITEM, SHCNF_PATH or SHCNF_FLUSHNOWAIT, PChar(FFileName), Nil);
      MessageBox(0, PChar('The magic number of ' + FFileName + ' has been set to ' + IntToStr(NumDropped)), 'DUTShellDropHandler', MB_OK OR MB_ICONINFORMATION);
      Result := S_OK;
   end; { else }
end; { IDropTarget.Drop }

initialization
   TComObjectFactory.Create(COMServer, TDropHandler, CLSID_DelDrop, '', 'DUTShellDropHandler', ciMultiInstance);

end. { unit DUTShellDropHandler }
