unit DUTShellIconHandler;
{..$DEFINE Debug}
{..$DEFINE UseResource}
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   30 January 1999
   Updated 17 February 1999
    1999 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use the 
   component or any part of its code, you must include an 
   acknowledgement in the finished code or executable (in the About 
   box and/or the Help file).

   The shell IconHandler extension.

   IconHandlers are registered under the target file type and
   are invoked when the shell requires an icon representation
   of the target file.
   There can only be one DropHandler or IconHandler registered
   under any given target.

   The IExtractIcon interface controls which icon is used for
   a given file --it can supply a different icon for each file
   of a given type.
   This example shows a simple rectangular "gauge", with the
   gauge height depending on the magic number stored in the
   .DEL file (values of 30+ are represented as "full").

   If the UseResource directive is defined, this object will
   supply pre-compiled icons.  If not, it will create icons
   on the fly.

   When the shell is about display an icon for a file, it does the
   following:
   (1) Finds its ProgID and ClassID.
   (2) If the file has a ClassID, it gets the icon location string from the
     'DefaultIcon' key under it.  The string indicates either per-class
     icon (e.g., 'FOOBAR.DLL,2') or per-instance icon (e.g., '%1,1').
   (3) If a per-instance icon is specified, the shell creates an icon
     extraction handler object for it (through IPersistFile::Load),
     and extracts the icon from it.
   It is important to note that the shell calls IExtractIcon.GetIconLocation
   first, then calls IExtractIcon.Extract.  Most application programs
   that support per-instance icons will probably store an icon location
   (DLL/EXE name and index/id) rather than an icon image in each file.
   In those cases, a programmer needs to implement only the GetIconLocation
   member and the Extract member simply returns S_FALSE.
   The Extract member need only be implemented if the icon image is
   stored within the files themselves or some other database, or if the
   icon needs to be created on the fly from the file contents themselves
   (.BMP thumbnails are a perfect example).

   In this case,
   HKEY_CLASSES_ROOT\.del = DelShellFile
   HKEY_CLASSES_ROOT\DefaultIcon = %1
   HKEY_CLASSES_ROOT\DelShellFile\shellex\IconHandler = CLSID_DelIcon

   Note how we override the IPersistFile-related DoSave from TCOMObjExtractIcon's
   ancestor TCOMObjPersistFile as well as TCOMObjExtractIcon's Do... methods.
   We needed to do this as both of these ancestors needed to be customized.
*)
interface

uses
   Windows,
{$IFNDEF UseResource}
   Graphics, { Graphics.TBitmap, etc. }
{$ENDIF}
   COMObj, { TComObjectFactory }
   COMObjExtractIcon, { TCOMObjExtractIcon }
   ShlObj; { IExtractIcon }

const
   CLSID_DelIcon : TGUID = (
      D1:$6A5EF9E4; D2:$DFCE; D3:$11CF; D4:($9A, $34, $44, $45, $53, $54, $00, $00));

type
   TIconHandler = class(TCOMObjExtractIcon)
   protected
      //IPersistFile
      function DoSave(FileName : String; Remember : BOOL) : HResult; override;
      //IExtractIcon
      function DoGetIconLocation(IconFileLength, IconFlags : UINT; var IconFile : String; var Index, Flags : UINT) : HRESULT; override;
      function DoExtractIcon(IconFile : String; Index : UINT; SizeBig, SizeSmall : Word; var hIconBig, hIconSmall : hIcon) : HRESULT; override;
   end;

implementation

uses
{$IFDEF Debug}
DLLDebug, { Loggit }
{$ENDIF}
{$IFNDEF UseResource}
   Classes, { Rect }
   SysUtils, { PByteArray }
{$ENDIF}
   COMServ, { COMServer }
   DUTDelShellUtils;

{$IFDEF UseResource}
{$R DUTShellIconHandler.RES}
{$ENDIF}

{ ########## TIconHandler ########## }

function TIconHandler.DoSave(FileName : String; Remember : BOOL) : HResult;
var
   Num : Integer;
begin
{$IFDEF Debug}
Loggit('TIconHandler.DoSave Begins');
{$ENDIF}
   Result := inherited DoSave(FileName, Remember);
   if FileName <> '' then begin
      //Save the object under a different filename.
      Result := E_FAIL;
      Num := GetDelFileNumber(FFileName);
      if Num < 0 then Exit;
      if SetDelFileNumber(Filename, Num) < 0 then Exit;
      SHChangeNotify(SHCNE_UPDATEITEM, SHCNF_PATH or SHCNF_FLUSHNOWAIT, PChar(FileName), Nil);
   end; { then }
   { else
     we were asked to save the "dirty" object under
     its current FFileName; no action necessary.
   }
{$IFDEF Debug}
Loggit('TIconHandler.DoSave Ends');
{$ENDIF}
end; { TCOMObjPersistFile.DoSave }

function TIconHandler.DoGetIconLocation(IconFileLength, IconFlags : UINT; var IconFile : String; var Index, Flags : UINT) : HRESULT;
{$IFDEF UseResource}
var
   DFN : Integer;
begin
{$IFDEF Debug}
Loggit('TIconHandler.DoGetIconLocation Begins');
{$ENDIF}
   //Simply returns S_FALSE
   Result := inherited DoGetIconLocation(IconFileLength, IconFlags, IconFile, Index, Flags);
   DFN := GetDelFileNumber(FFileName);
   case DFN of
   0..44      :
      Index := (DFN div 5) + 1; { 1..9 }
   45..MaxInt :
      Index := 9;
   else
      Index := 0;
   end; { case }
   SetLength(IconFile, IconFileLength);
   //Set IconFile to this DLL's path and name
   GetModuleFilename(hInstance, PChar(IconFile), IconFileLength);
   Flags := GIL_PERINSTANCE or GIL_DONTCACHE;
{$ELSE}
begin
{$IFDEF Debug}
Loggit('TIconHandler.DoGetIconLocation Begins');
{$ENDIF}
   //Simply returns S_FALSE
   inherited DoGetIconLocation(IconFileLength, IconFlags, IconFile, Index, Flags);

   Index := GetDelFileNumber(FFileName);
   //IconFile remains undefined
   Flags := GIL_PERINSTANCE or GIL_DONTCACHE or GIL_NOTFILENAME;
{$ENDIF}
   Result := S_OK;
{$IFDEF Debug}
Loggit('TIconHandler.DoGetIconLocation Ends');
{$ENDIF}
end; { IExtractIcon.DoGetIconLocation }

function TIconHandler.DoExtractIcon(IconFile : String; Index : UINT; SizeBig, SizeSmall : Word; var hIconBig, hIconSmall : hIcon) : HRESULT;
{$IFNDEF UseResource}
   function IconGraph(dim, dfNum : Integer) : hIcon;
   {
      Creates a size dim icon appropriate for magic number dfNum.
   }
   var
      vTB     : Windows.TBitmap;
      PictBmp,
      MaskBmp : Graphics.TBitmap;
      PictSiz,
      MaskSiz : Integer;
      NewPict,
      NewMask : PByteArray;
   begin
      //Return a null hIcon in case of failure
      Result := 0;
      //Create working image and mask
      PictBmp := TBitmap.Create;
      MaskBmp := TBitmap.Create;
      try
         with PictBmp do begin
            Width  := dim;
            Height := dim;
            Canvas.Brush.Color := clBlack;
            Canvas.Pen.Color := clBlack;
            Canvas.FillRect(Rect(0, 0, dim, dim));
            GetObject(Handle, SizeOf(vTB), @vTB);
            PictSiz := (Height * Width * vTB.bmBitsPixel * vTB.bmPlanes) div 8;
            MaskSiz := (Height * Width) div 8; //Mask is always two-color
            Canvas.Pen.Color   := clBlue;
            Canvas.Brush.Color := clWhite;
            Canvas.Rectangle(dim div 4, 0, 3*(dim div 4), dim);
            if dfNum < 0 then begin
               {bad value}
               Canvas.MoveTo(dim div 4, 0);
               Canvas.LineTo(3*(dim div 4) - 1, dim - 1);
               Canvas.MoveTo(dim div 4, dim-1);
               Canvas.LineTo(3*(dim div 4) - 1, 0);
            end else if dfNum >= 30 then begin
               {full}
               Canvas.Brush.Color := clRed;
               Canvas.FillRect(Rect((dim div 4) + 1, 1,
                                    3*(dim div 4) - 1, dim - 1));
            end else begin
               Canvas.Brush.Color := clRed;
               Canvas.FillRect(Rect((dim div 4) + 1, (dim - 1) - (dfNum*(dim - 2) div 30),
                                    3*(dim div 4) - 1, dim - 1));
            end; { else }
         end; { with }
         with MaskBmp do begin
            Width := dim;
            Height := dim;
            Canvas.Brush.Color := clWhite;
            Canvas.FillRect(Rect(0, 0, dim, dim));
            Monochrome := True;
            Canvas.Brush.Color := clBlack;
            Canvas.FillRect(Rect(dim div 4, 0, 3*(dim div 4), dim));
         end; { with }
         GetMem(NewPict, PictSiz);
         GetMem(NewMask, MaskSiz);
         try
            GetBitmapBits(PictBmp.Handle, PictSiz, NewPict);
            GetBitmapBits(MaskBmp.Handle, MaskSiz, NewMask);
            Result := CreateIcon(hInstance, vTB.bmWidth, vTB.bmHeight, vTB.bmPlanes, vTB.bmBitsPixel, NewMask, NewPict);
         finally
            FreeMem(NewPict, PictSiz);
            FreeMem(NewMask, MaskSiz);
         end; { try }
      finally
         PictBmp.Free;
         MaskBmp.Free;
      end; { try }
   end; { IconGraph }
{$ENDIF}
begin
{$IFDEF Debug}
Loggit('TIconHandler.DoExtractIcon Begins');
{$ENDIF}
   //Force caller to follow path to icon resource and extract the icon from there.
   //Simply returns S_FALSE
   Result := inherited DoExtractIcon(IconFile, Index, SizeBig, SizeSmall, hIconBig, hIconSmall);
{$IFNDEF UseResource}
   hIconBig := IconGraph(SizeBig, Index);
   hIconSmall := IconGraph(SizeSmall, Index);
   if (hIconBig <> 0) and (hIconSmall <> 0) then begin
      Result := NOERROR;
   end; { then }
{$ENDIF}
{$IFDEF Debug}
Loggit('TIconHandler.DoExtractIcon Ends');
{$ENDIF}
end; { IExtractIcon.DoExtractIcon }

initialization
   TComObjectFactory.Create(COMServer, TIconHandler, CLSID_DelIcon, '', 'DUTShellIconHandler', ciMultiInstance);

end. { unit DUTShellIconHandler }
